// Copyright (c) 1999-2000  David Muse
// See the file COPYING for more information

#include <config.h>
#include <defaults.h>
#include <datatypes.h>
#include <defines.h>
#include <sqlrclient.h>
#include <fstream.h>

#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
extern "C" {
	#undef __cplusplus
	#include <fcntl.h>
	#define __cplusplus
}
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif

#ifdef __GNUC__
static	int	sqlrclient::cachefilesequence;
#endif

sqlrclient::sqlrclient(char *server, int port, char *socket,
					char *user, char *password, 
					int retrytime, int tries) 
: client() {

	// initialize...

	// connection
	this->server=server;
	listenerinetport=(unsigned short int)port;
	listenerunixport=socket;
	this->retrytime=retrytime;
	this->tries=tries;

	// authentication
	this->user=user;
	this->password=password;
	userlen=strlen(user);
	passwordlen=strlen(password);

	// session state
	cached=0;
	connected=0;
	clearSessionFlags();

	// query
	querybuffer=NULL;
	fullpath=NULL;

	// result set
	rsbuffersize=0;

	firstrowindex=0;
	rowcount=0;
	totalrows=0;
	affectedrows=0;
	endofresultset=1;

	colcount=0;
	error=NULL;

	firstrow=NULL;
	firstcol=NULL;

	columnarray=NULL;
	columnnamearray=NULL;

	fieldarray=NULL;

	returnnulls=0;

	// cache file
	cachesourcefd=-1;
	cachedestname=NULL;
	cachedestfd=NULL;
	cacheon=0;

	// database id
	id=NULL;

	// debug off
	debug=0;

	// debug print function
	printfunction=NULL;

	// copy references flag
	copyrefs=0;
}

sqlrclient::~sqlrclient() {

	// unless it was already ended or suspended, end the session
	if (!endsessionsent && !suspendsessionsent) {
		endSession();
	}

	if (cachedestname) {
		delete cachedestname;
	}

	if (id) {
		delete[] id;
	}

	if (copyrefs) {
		deleteVariables();
		delete[] server;
		delete[] listenerunixport;
		delete[] user;
		delete[] password;
	}

	if (querybuffer) {
		delete[] querybuffer;
	}

	if (fullpath) {
		delete[] fullpath;
	}

	clearResultSet();
}

void	sqlrclient::clearSessionFlags() {
	endsessionsent=0;
	suspendsessionsent=0;
}

void	sqlrclient::clearVariables() {
	subcount=0;
	inbindcount=0;
	outbindcount=0;
}

void	sqlrclient::initVariables() {
	for (int i=0; i<MAXVAR; i++) {
		inbindvars[i].variable=NULL;
		inbindvars[i].value.stringval=NULL;
		inbindvars[i].type=STRING;
		outbindvars[i].variable=NULL;
		outbindvars[i].value.stringval=NULL;
		outbindvars[i].type=STRING;
		subvars[i].variable=NULL;
		subvars[i].value.stringval=NULL;
		subvars[i].type=STRING;
	}
}

void	sqlrclient::deleteVariables() {

	// if we were copying values, delete them
	for (int i=0; i<MAXVAR; i++) {
		if (inbindvars[i].variable) {
			delete[] inbindvars[i].variable;
		}
		if (inbindvars[i].type==STRING && 
			inbindvars[i].value.stringval) {
			delete[] inbindvars[i].value.stringval;
		}
		if (outbindvars[i].variable) {
			delete[] outbindvars[i].variable;
		}
		if (outbindvars[i].type==STRING && 
			outbindvars[i].value.stringval) {
			delete[] outbindvars[i].value.stringval;
		}
		if (subvars[i].variable) {
			delete[] subvars[i].variable;
		}
		if (subvars[i].type==STRING && 
			subvars[i].value.stringval) {
			delete[] subvars[i].value.stringval;
		}
	}
}

void	sqlrclient::clearResultSet() {

	clearCacheDest();
	clearCacheSource();
	clearError();
	clearColumns();

	// clear row counters
	firstrowindex=0;
	rowcount=0;
	totalrows=0;
	affectedrows=0;
	endofresultset=0;

	clearRowBuffers();
}

void	sqlrclient::clearCacheDest() {
	if (cachedestfd) {
		cachedestfd->close();
		delete cachedestfd;
		cachedestfd=NULL;
	}
}

void	sqlrclient::clearCacheSource() {
	if (cachesourcefd>-1) {
		close(cachesourcefd);
		cachesourcefd=-1;
	}
}

void	sqlrclient::clearError() {

	if (error) {
		delete error;
		error=NULL;
	}
}

void	sqlrclient::clearRowBuffers() {

	// delete linked list storing result set fields
	if (firstrow) {
		currentrow=firstrow;
		while (currentrow) {
			firstrow=currentrow->next;
			delete currentrow;
			currentrow=firstrow;
		}
		firstrow=NULL;
	}
	currentrow=NULL;

	// delete 2d array pointing to each field in the list
	if (fieldarray) {
		int i=0;
		while (fieldarray[i]) {
			delete[] fieldarray[i];
			i++;
		}
		delete[] fieldarray;
		fieldarray=NULL;
	}
}

void	sqlrclient::clearColumns() {

	// reset the counter to 0
	colcount=0;

	// delete linked list storing the column descriptions
	if (firstcol) {
		currentcol=firstcol;
		while (currentcol) {
			firstcol=currentcol->next;
			delete currentcol;
			currentcol=firstcol;
		}
		firstcol=NULL;
	}
	currentcol=NULL;

	// delete array pointing to each column
	if (columnarray) {
		delete[] columnarray;
		columnarray=NULL;
	}

	// delete array pointing to each column name
	if (columnnamearray) {
		delete[] columnnamearray;
		columnnamearray=NULL;
	}
}

int	sqlrclient::resumeSession(int port, char *socket) {
	return resumeCachedSession(port,socket,NULL);
}

int	sqlrclient::resumeCachedSession(int port, char *socket, 
						char *cachefile) {

	// flag this session as a resumed session
	resumed=1;
	
	// if already connected, end the session
	if (connected) {
		endSession();
	}

	// handle the cachefile
	if (cachefile && cachefile[0]) {

		// clean up old cachefilee
		if (cachedestname) {
			delete cachedestname;
		}
	
		// create new cachedestname
		cachedestname=new strstream();
		*cachedestname << cachefile << ends;
	}

	// set the connectionunixport and connectioninetport values
	connectionunixport=socket;
	connectioninetport=(unsigned short int)port;

	// first, try for the unix port
	if (socket && socket[0]) {
		connected=openUnixConnection(socket,retrytime,tries);
	}

	// then try for the inet port
	if (!connected) {
		connected=openInetConnection(server,port,retrytime,tries);
	}

	if (debug) {
		debugPrint("Resuming Session: ");
	}

	if (connected) {

		if (debug) {
			debugPrint("success");
		       	debugPrint("\n");
		}

		// initialize session flags
		clearSessionFlags();

		// The server will send the index of the last row from the
		// previous result set if it was suspended. If it wasn't
		// suspended, it will send a -1.  Initialize firstrowindex
		// and rowcount from this index. (if firstrowindex 
		// is > -1 then it will be overridden in parseData below)
		readFromServer(&firstrowindex);
		rowcount=firstrowindex+1;

		// if the previous result set was suspended, then we need
		// to pick up some rows from it
		if (firstrowindex>-1) {

			if (debug) {
				debugPrint("Previous result set was ");
			       	debugPrint("suspended at row index: ");
				debugPrint((long)firstrowindex);
				debugPrint("\n");
			}

			// clearResultSet() basically...
			// without resetting rowcount or firstrowindex 
			// or the cache file
			clearError();
			clearColumns();
			endofresultset=0;
			clearRowBuffers();

			// process the result set
			if (cachefile && cachefile[0]) {
				cacheOn();
			}
			if (rsbuffersize) {
				processResultSet(firstrowindex+
							rsbuffersize-1,0);
			} else {
				processResultSet();
			}
		}

	} else {
		if (debug) {
			debugPrint("failure");
			debugPrint("\n");
		}
	}

	return connected;
}

void	sqlrclient::setResultSetBufferSize(int rows) {
	rsbuffersize=rows;
}

int	sqlrclient::getResultSetBufferSize() {
	return	rsbuffersize;
}

int	sqlrclient::ping() {

	validatebinds=0;
	resumed=0;
	if (sendQueryInternal("ping")) {

		// fake input/output binds
		writeToServer((short)0);
		writeToServer((short)0);

		// get the ping result
		short	result;
		getShort(&result);

		endofresultset=1;

		return (int)result;
	}
	return 0;
}

char	*sqlrclient::identify() {

	validatebinds=0;
	resumed=0;
	if (sendQueryInternal("identify")) {

		// fake input/output binds
		writeToServer((short)0);
		writeToServer((short)0);

		// get the id
		short	size;
		getShort(&size);
		if (id) {
			delete[] id;
		}
		id=new char[size+1];
		getString(id,size);
		id[size]=(char)NULL;

		endofresultset=1;

		return id;
	}
	return 0;
}

void	sqlrclient::cacheOn() {
	cacheon=1;
	cachettl=600;
}

void	sqlrclient::setCacheTtl(int ttl) {
	cachettl=ttl;
}

void	sqlrclient::cacheOff() {
	cacheon=0;
}

int	sqlrclient::sendQuery(char *query) {
	prepareQuery(query);
	return executeQuery();
}

int	sqlrclient::sendFileQuery(char *path, char *filename) {
	prepareFileQuery(path,filename);
	return executeQuery();
}

void	sqlrclient::initQueryBuffer() {
	if (!querybuffer) {
		querybuffer=new char[MAXQUERYSIZE+1];
		queryptr=querybuffer;
	}
}

void	sqlrclient::prepareQuery(char *query) {
	validatebinds=0;
	resumed=0;
	clearVariables();
	if (copyrefs) {
		initQueryBuffer();
		strcpy(queryptr,query);
	} else {
		queryptr=query;
	}
}

void	sqlrclient::prepareFileQuery(char *path, char *filename) {

	// init some variables
	validatebinds=0;
	resumed=0;
	clearVariables();

	// init the fullpath buffer
	if (!fullpath) {
		fullpath=new char[MAXPATHLEN+1];
	}

	// add the path to the fullpath
	int	index=0;
	int	counter=0;
	while (path[index] && counter<MAXPATHLEN) {
		fullpath[counter]=path[index];
		index++;
		counter++;
	}

	// add the "/" to the fullpath
	if (counter<=MAXPATHLEN) {
		fullpath[counter]='/';
		counter++;
	}

	// add the file to the fullpath
	index=0;
	while (filename[index] && counter<MAXPATHLEN) {
		fullpath[counter]=filename[index];
		index++;
		counter++;
	}

	// handle a filename that's too long
	if (counter>MAXPATHLEN) {

		// sabotage the file name so it can't be opened
		fullpath[0]=(char)NULL;

		// debug info
		if (debug) {
			debugPrint("File name ");
			debugPrint(path);
			debugPrint("/");
			debugPrint(filename);
			debugPrint(" is too long.");
			debugPrint("\n");
		}

	} else {

		// terminate the string
		fullpath[counter]=(char)NULL;

		// debug info
		if (debug) {
			debugPrint("File: ");
			debugPrint(fullpath);
			debugPrint("\n");
		}
	}

	// open the file
	ifstream queryfile(fullpath);

	// handle a file we couldn't open
	if (!queryfile.good()) {
		debugPrint("The file ");
		debugPrint(fullpath);
		debugPrint(" could not be opened.");
		debugPrint("\n");
		return 0;
	}

	initQueryBuffer();

	// read the file into the query buffer
	index=0;
	for (;;) {
		if (index==MAXQUERYSIZE) {
			debugPrint("The query in ");
			debugPrint(fullpath);
			debugPrint(" is too large.");
			debugPrint("\n");
			break;
		}
		querybuffer[index]=queryfile.rdbuf()->sgetc();
		if (querybuffer[index]<=0) {
			break;
		}
		queryfile.rdbuf()->snextc();
		index++;
	}
	querybuffer[index]=NULL;

	queryfile.close();
}

void	sqlrclient::substitution(char *variable, char *value) {
	if (subcount<MAXVAR && variable && variable[0]) {
		stringVar(&subvars[subcount],variable,value);
		subcount++;
	}
}

void	sqlrclient::substitution(char *variable, long value) {
	if (subcount<MAXVAR && variable && variable[0]) {
		longVar(&subvars[subcount],variable,value);
		subcount++;
	}
}

void	sqlrclient::substitution(char *variable, double value, 
					short precision, short scale) {
	if (subcount<MAXVAR && variable && variable[0]) {
		doubleVar(&subvars[subcount],variable,value,precision,scale);
		subcount++;
	}
}

void	sqlrclient::inputBind(char *variable, char *value) {
	if (inbindcount<MAXVAR && variable && variable[0]) {
		stringVar(&inbindvars[inbindcount],variable,value);
		inbindvars[inbindcount].send=1;
		inbindcount++;
	}
}

void	sqlrclient::inputBind(char *variable, long value) {
	if (inbindcount<MAXVAR && variable && variable[0]) {
		longVar(&inbindvars[inbindcount],variable,value);
		inbindvars[inbindcount].send=1;
		inbindcount++;
	}
}

void	sqlrclient::inputBind(char *variable, double value, 
				short precision, short scale) {
	if (inbindcount<MAXVAR && variable && variable[0]) {
		doubleVar(&inbindvars[inbindcount],variable,value,
						precision, scale);
		inbindvars[inbindcount].send=1;
		inbindcount++;
	}
}

void	sqlrclient::substitutions(char **variables, char **values) {
	int	index=0;
	while (variables[index] && subcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			stringVar(&subvars[subcount],
					variables[index],values[index]);
			subcount++;
		}
		index++;
	}
}

void	sqlrclient::substitutions(char **variables, long *values) {
	int	index=0;
	while (variables[index] && subcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			longVar(&subvars[subcount],
					variables[index],values[index]);
			subcount++;
		}
		index++;
	}
}

void	sqlrclient::substitutions(char **variables, double *values, 
					short *precisions, short *scales) {
	int	index=0;
	while (variables[index] && subcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			doubleVar(&subvars[subcount],
					variables[index],
					values[index],
					precisions[index],
					scales[index]);
			subcount++;
		}
		index++;
	}
}

void	sqlrclient::inputBinds(char **variables, char **values) {
	int	index=0;
	while (variables[index] && inbindcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			stringVar(&inbindvars[inbindcount],
					variables[index],values[index]);
			inbindvars[inbindcount].send=1;
			inbindcount++;
		}
		index++;
	}
}

void	sqlrclient::inputBinds(char **variables, long *values) {
	int	index=0;
	while (variables[index] && inbindcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			longVar(&inbindvars[inbindcount],
					variables[index],values[index]);
			inbindvars[inbindcount].send=1;
			inbindcount++;
		}
		index++;
	}
}

void	sqlrclient::inputBinds(char **variables, double *values, 
						short *precisions,
						short *scales) {
	int	index=0;
	while (variables[index] && inbindcount<MAXVAR) {
		if (variables[index] && variables[index][0]) {
			doubleVar(&inbindvars[inbindcount],
					variables[index],
					values[index],
					precisions[index],
					scales[index]);
			inbindvars[inbindcount].send=1;
			inbindcount++;
		}
		index++;
	}
}

void	sqlrclient::stringVar(bindvar *var, char *variable, char *value) {

	initVar(var,variable);
	var->type=STRING;

	// store the value, handle NULL values too
	if (value) {
		if (copyrefs) {
			var->value.stringval=new char[strlen(value)+1];
			strcpy(var->value.stringval,value);
		} else {
			var->value.stringval=value;
		}
		var->valuesize=strlen(value);
	} else {
		var->value.stringval=NULL;
		var->valuesize=NULL_VALUE_SIZE;
	}
}

void	sqlrclient::longVar(bindvar *var, char *variable, long value) {
	initVar(var,variable);
	var->type=LONG;
	var->value.longval=value;
}

void	sqlrclient::doubleVar(bindvar *var, char *variable, double value,
					short precision, short scale) {
	initVar(var,variable);
	var->type=DOUBLE;
	var->value.doubleval.value=value;
	var->value.doubleval.precision=precision;
	var->value.doubleval.scale=scale;
}

void	sqlrclient::initVar(bindvar *var, char *variable) {

	// clear any old variable name that was stored and assign the new 
	// variable name also clear any old value that was stored in this 
	// variable
	if (copyrefs) {
		if (var->variable) {
			delete[] var->variable;
		}
		var->variable=new char[strlen(variable)+1];
		strcpy(var->variable,variable);

		if (var->type==STRING &&
			var->value.stringval) {
			delete[] var->value.stringval;
		}
	} else {
		var->variable=variable;
	}
}

void	sqlrclient::defineOutputBind(char *variable, int length) {
	if (outbindcount<MAXVAR && variable && variable[0]) {
		if (copyrefs) {
			if (outbindvars[outbindcount].variable) {
				delete[] outbindvars[outbindcount].variable;
			}
			outbindvars[outbindcount].variable=new 
						char[strlen(variable)+1];
			strcpy(outbindvars[outbindcount].variable,variable);
			if (outbindvars[outbindcount].type==STRING &&
				outbindvars[outbindcount].value.stringval) {
				delete[] outbindvars[outbindcount].
							value.stringval;
			}
		} else {
			outbindvars[outbindcount].variable=variable;
		}
		outbindvars[outbindcount].value.stringval=NULL;
		outbindvars[outbindcount].valuesize=length;
		outbindvars[outbindcount].send=1;
		outbindcount++;
	}
}

char	*sqlrclient::getOutputBind(char *variable) {

	if (variable) {
		for (int i=0; i<outbindcount; i++) {
			if (!strcmp(outbindvars[i].variable,variable)) {
				return outbindvars[i].value.stringval;
			}
		}
	}
	return NULL;
}

void	sqlrclient::validateBinds() {
	validatebinds=1;
}

int	sqlrclient::executeQuery() {

	// a useful variable
	int	retval=0;

	if (!subcount) {

		// validate the bind variables
		if (validatebinds) {
			validateBindsInternal(queryptr);
		}
		
		// run the query
		retval=runQuery(queryptr);

	} else {

		// perform substitutions
		strstream	*container=new strstream();
		char		*ptr=queryptr;
		int		found=0;
		int		inquotes=0;
		int		inbraces=0;
		int		len=0;
		strstream	*braces;


		// iterate through the string
		while (*ptr) {
		
			// figure out whether we're inside a quoted 
			// string or not
			if (*ptr=='\'' && *(ptr-1)!='\\') {
				if (inquotes) {
					inquotes=0;
				} else {
					inquotes=1;
				}
			}
		
			// if we find an open-brace then start 
			// sending to a new buffer
			if (*ptr=='[' && !inbraces && !inquotes) {
				braces=new strstream();
				inbraces=1;
				ptr++;
			}
		
			// if we find a close-brace then process 
			// the brace buffer
			if (*ptr==']' && inbraces && !inquotes) {
				*braces << ends;
		
				// look for an = sign, skipping whitespace
				char	*bptr=braces->str();
				while (*bptr && (*bptr==' ' || 
					*bptr=='	' || *bptr=='\n')) {
					bptr++;
				}
		
				if (*bptr=='=') {
					// if we find an equals sign first 
					// then process the rest of the buffer
					bptr++;
		
					// skip whitespace
					while (*bptr && (*bptr==' ' || 
						*bptr=='	' || 
					 	*bptr=='\n')) {
						bptr++;
					}
		
					// if the remaining contents of the 
					// buffer are '' or nothing then we 
					// must have an ='' or just an = with 
					// some whitespace, replace this
					// with "is NULL" otherwise, just write
					// out the contents of the buffer
					if (!bptr || 
						(bptr && !strcmp(bptr,"''"))) {
						*container << " is NULL ";
					} else {
						*container << braces->str();
					}
				} else {
					// if we don't find an equals sign, 
					// then write the contents out directly
					*container << braces->str();
				}
				delete braces;
				inbraces=0;
				ptr++;
			}
		
			// if we encounter $(....) then replace the 
			// variable within
			if ((*ptr)=='$' && (*(ptr+1))=='(') {
		
				// first iterate through the arrays passed in
				found=0;
				for (int i=0; i<subcount && !found; i++) {
		
					// if we find a match, write the 
					// value to the container and skip 
					// past the $(variable)
					len=strlen(subvars[i].variable);
					if (!strncmp((ptr+2),
						subvars[i].variable,len) &&
						(*(ptr+2+len))==')') {
		
						if (subvars[i].type==STRING) {
							if (inbraces) {
								*braces << 
								subvars[i].
								value.stringval;
							} else {
								*container << 
								subvars[i].
								value.stringval;
							}
						} else 
						if (subvars[i].type==LONG) {
							if (inbraces) {
								*braces << 
								subvars[i].
								value.longval;
							} else {
								*container << 
								subvars[i].
								value.longval;
							}
						} else 
						if (subvars[i].type==DOUBLE) {
							if (inbraces) {
								*braces << 
								subvars[i].
								value.doubleval.
									value;
							} else {
								*container << 
								subvars[i].
								value.doubleval.
									value;
							}
						}
						ptr=ptr+3+len;
						found=1;
					}
				}
		
				// if the variable wasn't found, then 
				// just write the $(
				if (!found) {
					if (inbraces) {
						*braces << "$(";
					} else {
						*container << "$(";
					}
					ptr=ptr+2;
				}
		
			} else {
		
				// print out the current character and proceed
				if (inbraces) {
					*braces << *ptr;
				} else {
					*container << *ptr;
				}
				ptr++;
			}
		}
		*container << ends;

		// validate the bind variables
		if (validatebinds) {
			validateBindsInternal(container->str());
		}

		// run the query
		retval=runQuery(container->str());

		// clean up
		delete container;
	}

	return retval;
}

int	sqlrclient::runQuery(char *query) {

	// send the query
	if (sendQueryInternal(query)) {

		sendInputBinds();
		sendOutputBinds();

		if (processResultSet()) {
			return 1;
		}
	}
	return 0;
}

char	*sqlrclient::getCacheFileName() {
	if (cachedestname) {
		return	cachedestname->str();
	}
	return NULL;
}

int	sqlrclient::openSession() {

	if (debug) {
		debugPrint("Connecting to listener...");
		debugPrint("\n");
	}

	// open a connection to the listener
	int	openresult=0;

	// first, try for a unix connection
	if (listenerunixport && listenerunixport[0]) {
		openresult=openUnixConnection(listenerunixport,retrytime,tries);
	}

	// then try for an inet connection
	if (!openresult && listenerinetport) {
		openresult=openInetConnection(server,listenerinetport,
						retrytime,tries);
	}

	if (openresult) {

		// authenticate with the listner
		handleAuthentication();

		// try to get the port of the connection daemon
		// to talk to.
		int	success=getNewPort();

		// close the connection to the listener
		client::closeConnection();

		// handle failure to get a new port
		if (!success) {
			return 0;
		}

		if (debug) {
			debugPrint("Connecting to connection...");
			debugPrint("\n");
		}

		// first, try for the unix port
		if (listenerunixport && listenerunixport[0] &&
						connectionunixport) {
			connected=openUnixConnection(connectionunixport,
							retrytime,tries);
		}
 
		// then try for the inet port
		if (!connected && connectioninetport) {
			connected=openInetConnection(server,connectioninetport,
							retrytime,tries);
		}

		if (connected) {

			if (debug) {
				debugPrint("Connected.");
				debugPrint("\n");
			}

			clearSessionFlags();

			// The server will send the index of the last row from 
			// the previous result set if it was suspended, if not,
			// it will send a -1.
			// In this case (a new connection) it will always be -1
			// but we have to pick it up anyway.
			readFromServer(&firstrowindex);
		}

		return connected;
	}

	// handle failure to connect, but don't override a pre-existing error
	if (!error) {
		error=new char[42];
		strcpy(error,"Couldn't connect to SQL Relay connection.");
	}
	return 0;
}

void	sqlrclient::closeConnection() {

	if (connected) {
		client::closeConnection();
		connected=0;
	}
}

int	sqlrclient::openCachedResultSet(char *filename) {
	return openCachedResultSet(filename,-1,0);
}

int	sqlrclient::openCachedResultSet(char *filename, int offset, int row) {

	if (debug) {
		debugPrint("Opening cached result set: ");
		debugPrint(filename);
		debugPrint("\n");
		if (row>-1) {
			debugPrint("Setting first row to: ");
			debugPrint((long)row);
			debugPrint("\n");
		}
	}

	if (!endofresultset) {
		abortResultSet();
	}
	clearResultSet();

	cached=1;

	// open the file
	if ((cachesourcefd=open(filename,O_RDWR|O_EXCL))>-1) {

		// initialize firstrowindex and rowcount
		firstrowindex=row;
		rowcount=firstrowindex;

		// skip the ttl
		long	longvar;
		getLong(&longvar);

		// process the result set
		processResultSet(firstrowindex+rsbuffersize-1,offset);
		return 1;
	}
	return 0;
}

int	sqlrclient::endSession() {

	if (debug) {
		debugPrint("Ending Session");
		debugPrint("\n");
	}

	if (!endofresultset) {
		abortResultSet();
	}

	// write a ~ to the connection
	int	retval=1;
	if (connected) {
		retval=0;
		writeToServer((long)1);
		writeToServer("~",1);
		endsessionsent=1;
		retval=1;
		closeConnection();
	}
	return retval;
}

int	sqlrclient::suspendSession() {

	if (!connected && !openSession()) {
		return 0;
	}

	suspendResultSet();

	if (debug) {
		debugPrint("Suspending Session");
		debugPrint("\n");
	}

	// write a ! to the connection
	int	retval=0;
	writeToServer((long)1);
	writeToServer("!",1);
	suspendsessionsent=1;
	retval=1;
	closeConnection();

	return retval;
}

int	sqlrclient::getConnectionPort() {

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPrint("Getting connection port: ");
		debugPrint((long)connectioninetport);
		debugPrint("\n");
	}

	return (int)connectioninetport;
}

char	*sqlrclient::getConnectionSocket() {

	if (!connected && !openSession()) {
		return NULL;
	}

	if (debug) {
		debugPrint("Getting connection socket: ");
		if (connectionunixport) {
			debugPrint(connectionunixport);
		}
		debugPrint("\n");
	}

	if (connectionunixport) {
		return connectionunixport;
	}
	return NULL;
}

int	sqlrclient::getCacheFileOffset() {
	if (cachesourcefd>-1) {
		return lseek(cachesourcefd,(off_t)NULL,SEEK_CUR);
	}
	return 0;
}

void	sqlrclient::handleAuthentication() {

	if (debug) {
		debugPrint("Authenticating with : ");
		debugPrint(user);
		debugPrint(":");
		debugPrint(password);
		debugPrint("\n");
	}

	writeToServer((long)userlen);
	writeToServer(user,userlen);

	writeToServer((long)passwordlen);
	writeToServer(password,passwordlen);
}

int	sqlrclient::getNewPort() {

	// get the size of the unix port string
	short		size;
	getShort(&size);

	if (size==-2) {

		// if size is -2 then there has been an authentication error
		// clear any result set left over from a previous (sucessful) 
		// query
		clearResultSet();
		handleError(NULL);
		return 0;

	} else if (size<=MAXPATHLEN) {

		// get the unix port string
		getString(connectionunixportbuffer,size);
		connectionunixportbuffer[size]=(char)NULL;
		connectionunixport=connectionunixportbuffer;

	} else {

		// if size is too big, display the error
		debugPrint("The pathname of the unix port was too long: ");
		debugPrint((long)size);
		debugPrint(" bytes.  The maximum size is ");
		debugPrint((long)MAXPATHLEN);
		debugPrint(" bytes.\n");
		return 0;

	}

	// get the inet port
	getShort((short *)&connectioninetport);

	if (debug) {
		debugPrint("Reconnecting to ");
		debugPrint("\n");
		debugPrint("unix port: ");
		debugPrint(connectionunixport);
		debugPrint("\n");
		debugPrint("or inet port: ");
		debugPrint((long)connectioninetport);
		debugPrint("\n");
	}

	return 1;
}

int	sqlrclient::sendQueryInternal(char *query) {

	// if the first 8 characters of the query are "-- debug" followed
	// by a return, then set debugging on
	if (!strncmp(query,"-- debug\n",9)) {
		debugOn();
	}

	if (!endofresultset) {
		abortResultSet();
	}
	clearResultSet();

	cached=0;

	if (!connected && !openSession()) {
		return 0;
	}

	if (debug) {
		debugPrint("Sending Query:");
		debugPrint("\n");
		debugPrint(query);
		debugPrint("\n");
	}

	// write the query to the server.
	int	querylen=strlen(query);
	writeToServer((long)querylen);
	writeToServer(query,querylen);

	return 1;
}

void	sqlrclient::validateBindsInternal(char *query) {

	// some useful variables
	char	*ptr;
	char	*start;
	char	*after;
	int	found;
	int	len;
	int	count;

	// check each input bind
	count=inbindcount;
	for (int i=0; i<count; i++) {

		found=0;
		start=query+1;
		len=strlen(inbindvars[i].variable);

		// there may be more than 1 match for the variable name as in
		// "select * from table where table_name=:table_name", both
		// table_name's would match, but only the second is a bind
		// variable
		while (ptr=strstr(start,inbindvars[i].variable)) {

			// for a match to be a bind variable, it must be 
			// preceded by a colon and can't be followed by an
			// alphabet character, number or underscore
			after=ptr+len;
			if (*(ptr-1)==':' && *after!='_' &&
				!(*(after)>='a' && *(after)<='z') &&
				!(*(after)>='A' && *(after)<='Z') &&
				!(*(after)>='0' && *(after)<='9')) {
				found=1;
				break;
			} else {
				// jump past this instance to look for the
				// next one
				start=ptr+len;
			}
		}

		if (!found) {
			inbindvars[i].send=0;
			inbindcount--;
		}
	}

	// check each output bind
	count=outbindcount;
	for (int i=0; i<count; i++) {

		found=0;
		start=query+1;
		len=strlen(outbindvars[i].variable);

		// there may be more than 1 match for the variable name as in
		// "select * from table where table_name=:table_name", both
		// table_name's would match, but only 1 is correct
		while (ptr=strstr(start,outbindvars[i].variable)) {

			// for a match to be a bind variable, it must be 
			// preceded by a colon and can't be followed by an
			// alphabet character, number or underscore
			after=ptr+len;
			if (*(ptr-1)==':' && *after!='_' &&
				!(*(after)>='a' && *(after)<='z') &&
				!(*(after)>='A' && *(after)<='Z') &&
				!(*(after)>='0' && *(after)<='9')) {
				found=1;
				break;
			} else {
				// jump past this instance to look for the
				// next one
				start=ptr+len;
			}
		}

		if (!found) {
			outbindvars[i].send=0;
			outbindcount--;
		}
	}
}

void	sqlrclient::sendInputBinds() {

	if (debug) {
		debugPrint("Input Bind Variables:");
		debugPrint("\n");
	}

	// write the input bind variables/values to the server.
	writeToServer((short)inbindcount);
	short	size;
	int	count=inbindcount;
	for (int i=0; i<count; i++) {

		// don't send anything if the send flag is turned off
		if (!inbindvars[i].send) {
			count++;
			continue;
		}

		// send the variable
		size=(short)strlen(inbindvars[i].variable);
		writeToServer((short)size);
		writeToServer(inbindvars[i].variable,(int)size);

		if (debug) {
			debugPrint(inbindvars[i].variable);
			debugPrint("(");
			debugPrint((long)size);
		}

		// send the type
		writeToServer((short)inbindvars[i].type);
	
		if (inbindvars[i].type==STRING) {
			// send the value
			writeToServer((short)inbindvars[i].valuesize);
			if (inbindvars[i].valuesize>0) {
				writeToServer(inbindvars[i].value.stringval,
					(int)inbindvars[i].valuesize);
			}

			if (debug) {
				debugPrint(":STRING)=");
				debugPrint(inbindvars[i].value.stringval);
				debugPrint("(");
				debugPrint((long)inbindvars[i].valuesize);
				debugPrint(")");
				debugPrint("\n");
			}
		} else if (inbindvars[i].type==LONG) {
			// send the value
			writeToServer((long)inbindvars[i].value.longval);

			if (debug) {
				debugPrint(":LONG)=");
				debugPrint(inbindvars[i].value.longval);
				debugPrint("\n");
			}
		} else if (inbindvars[i].type==DOUBLE) {
			// send the value
			writeToServer((double)inbindvars[i].value.
							doubleval.value);
			writeToServer((short)inbindvars[i].value.
							doubleval.precision);
			writeToServer((short)inbindvars[i].value.
							doubleval.scale);

			if (debug) {
				debugPrint(":DOUBLE)=");
				debugPrint(inbindvars[i].value.
							doubleval.value);
				debugPrint(":");
				debugPrint((long)inbindvars[i].value.
							doubleval.precision);
				debugPrint(",");
				debugPrint((long)inbindvars[i].value.
							doubleval.scale);
				debugPrint("\n");
			}
		}
	}
}

void	sqlrclient::sendOutputBinds() {

	if (debug) {
		debugPrint("Output Bind Variables:");
		debugPrint("\n");
	}

	// write the output bind variables to the server.
	writeToServer((short)outbindcount);
	short	size;
	int	count=outbindcount;
	for (int i=0; i<count; i++) {

		// don't send anything if the send flag is turned off
		if (!outbindvars[i].send) {
			count++;
			continue;
		}

		// send the variable and size that the buffer needs to be
		size=(short)strlen(outbindvars[i].variable);
		writeToServer((short)size);
		writeToServer(outbindvars[i].variable,(int)size);
		writeToServer((short)outbindvars[i].valuesize);

		if (debug) {
			debugPrint(outbindvars[i].variable);
			debugPrint("(");
			debugPrint((long)outbindvars[i].valuesize);
			debugPrint(")\n");
		}
	}
}

int	sqlrclient::getShort(short *integer) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1) {
		return read(cachesourcefd,(void *)integer,sizeof(short));
	} else {
		return readFromServer(integer);
	}
}

int	sqlrclient::getLong(long *integer) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1) {
		return read(cachesourcefd,(void *)integer,sizeof(long));
	} else {
		return readFromServer(integer);
	}
}

int	sqlrclient::getChar(char *character) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1) {
		return read(cachesourcefd,(void *)character,sizeof(char));
	} else {
		return readFromServer(character,sizeof(char));
	}
}

int	sqlrclient::getString(char *string, int size) {

	// if the result set is coming from a cache file, read from
	// the file, if not, read from the server
	if (cachesourcefd>-1) {
		return read(cachesourcefd,(void *)string,size);
	} else {
		return readFromServer(string,size);
	}
}

void	sqlrclient::abortResultSet() {

	// if the end of the previous result set was never reached, the
	// connection daemon is still expecting a skip and fetch size, send
	// -1 for skip and 0 for fetch to signal to it that we're done with 
	// that result set, also read the -1 that the connection sends back
	// indicating that the end of the result set has been reached
	if (debug) {
		debugPrint("Aborting Result Set");
		debugPrint("\n");
	}
	if (connected && !cached) {
		writeToServer((long)-1);
		writeToServer((long)0);
		long	integer;
		readFromServer(&integer);
	}
	finishCaching();
}

void	sqlrclient::suspendResultSet() {

	// if the end of the previous result set was never reached, the
	// connection daemon is still expecting a skip and fetch size, send
	// -2 for skip and 0 for fetch to signal it to suspend that result set,
	// also read the -1 that the connection sends back indicating that the 
	// end of the result set has been reached
	if (!endofresultset) {
		if (debug) {
			debugPrint("Suspending Result Set");
			debugPrint("\n");
		}
		if (connected && !cached) {
			writeToServer((long)-2);
			writeToServer((long)0);
			long	integer;
			readFromServer(&integer);
		}
		suspendCaching();
	}
}

int	sqlrclient::processResultSet() {
	return processResultSet(rsbuffersize-1,0);
}

int	sqlrclient::processResultSet(int rowtoget, int offset) {

	// start caching the result set
	if (cacheon) {
		startCaching();
	}

	// parse the header and data
	int	success;
	if ((success=parseHeader())==1) {
		if (success=parseOutputBinds()==1) {
			success=parseData(rowtoget,offset);
		}
	}

	// handle error responses
	if (success==0) {
		handleError(NULL);
		return 0;
	} else if (success==-1) {
		handleError("Failed to execute the query and/or process the result set.\n A query, bind variable or bind value could be too large, there could be too \n many bind variables, or a network error may have ocurred.");
		closeConnection();
		return 0;
	}
	return 1;
}

void	sqlrclient::startCaching() {

	if (!resumed) {

		// clean up old cachefile name
		if (cachedestname) {
			delete cachedestname;
		}
	
		// create new cachefile name
		cachedestname=new strstream();
		*cachedestname << CACHE_DIR << "/" << getpid();
		*cachedestname << "-" << cachefilesequence << ends;
		cachefilesequence++;

		if (debug) {
			debugPrint("Caching data to ");
			debugPrint(cachedestname->str());
			debugPrint("\n");
		}

	} else {
		if (debug) {
			debugPrint("Resuming caching data to ");
			debugPrint(cachedestname->str());
			debugPrint("\n");
		}
	}

	// create the cachefile, truncate it unless we're 
	// resuming a previous session
	if (!resumed) {
		cachedestfd=new ofstream(cachedestname->str(), 
				ios::in | ios::out | ios::trunc);
	} else {
		cachedestfd=new ofstream(cachedestname->str(), 
				ios::in | ios::out | ios::ate | ios::app);
	}

	if (cachedestfd && cachedestfd->good()) {

		// write ttl to file
		if (!resumed) {
			long	expiration=time(NULL)+cachettl;
			cachedestfd->write((char *)&expiration,sizeof(long));
		}

	} else {

		if (debug) {
			debugPrint("Error caching data to ");
			debugPrint(cachedestname->str());
			debugPrint("\n");
		}

		// in case of an error, clean up
		clearCacheDest();
	}
}


void	sqlrclient::finishCaching() {

	if (!cachedestfd) {
		return;
	}

	// terminate the result set with a -1
	long	minusone=-1;
	cachedestfd->write((char *)&minusone,sizeof(long));

	// close the cache file and clean up
	clearCacheDest();
}

void	sqlrclient::suspendCaching() {

	if (!cachedestfd) {
		return ;
	}

	// close the cache file and clean up
	clearCacheDest();
}

int	sqlrclient::parseHeader() {

	if (debug) {
		debugPrint("Parsing Header");
		debugPrint("\n");
	}

	// first get the number of rows returned by the query
	if (getLong(&totalrows)<1) {
		return -1;
	}

	// if we get a -2, then there's an error
	if (totalrows==-2) {
		return 0;
	}

	// first get the number of rows affected by the query
	getLong(&affectedrows);

	if (debug) {
		debugPrint("Total row count: ");
		debugPrint(totalrows);
		debugPrint("\n");
		debugPrint("Affected row count: ");
		debugPrint(affectedrows);
		debugPrint("\n");
	}

	// a useful variable
	short	length;

	// get the header segment
	for (;;) {

		// get the column name length
		if (getShort(&length)!=sizeof(short)) {
			return -1;
		}

		// check for end of header
		if (!length) {
			break;
		}

		// new column
		if (!firstcol) {
			firstcol=new column();
			currentcol=firstcol;
		} else {
			currentcol->next=new column();
			currentcol=currentcol->next;
		}

		// get the column name
		colcount++;
		currentcol->name=new char[length+1];
		getString(currentcol->name,length);
		currentcol->name[length]=(char)NULL;

		// get the column type
		getShort(&currentcol->type);

		// get the column length
		getLong(&currentcol->length);

		if (debug) {
			debugPrint("\"");
			debugPrint(currentcol->name);
			debugPrint("\",");
			debugPrint("\"");
			debugPrint(datatypestring[currentcol->type]);
			debugPrint("\",");
			debugPrint("\"");
			debugPrint((long)currentcol->length);
			debugPrint("\"");
			debugPrint("\n");
		}
	}

	// terminate the column list
	if (currentcol) {
		currentcol->next=NULL;
	}

	createColumnArrays();

	// cache the column definitions
	cacheHeader();

	return 1;
}

int	sqlrclient::parseOutputBinds() {

	if (debug) {
		debugPrint("Receiving Output Bind Values: ");
		debugPrint("\n");
	}

	// useful variables
	short	length;
	int	count=0;

	// get the bind values
	for (;;) {

		// get the value length
		if (getShort(&length)!=sizeof(short)) {
			return -1;
		}

		// check for end of bind values
		if (length==-1) {
			break;
		}

		// get the value
		if (length==NULL_VALUE_SIZE) {
			if (returnnulls) {
				outbindvars[count].value.stringval=NULL;
			} else {
				outbindvars[count].value.stringval=new char[1];
				outbindvars[count].value.stringval[0]=
								(char)NULL;
			}
		} else {
			outbindvars[count].value.stringval=new char[length+1];
			getString(outbindvars[count].value.stringval,length);
			outbindvars[count].value.stringval[length]=(char)NULL;
		}

		if (debug) {
			debugPrint(outbindvars[count].variable);
			debugPrint("=");
			debugPrint(outbindvars[count].value.stringval);
			debugPrint("\n");
		}

		count++;
	}

	// cache the output binds
	cacheOutputBinds(count);

	return 1;
}

void	sqlrclient::cacheHeader() {

	if (resumed || !cachedestfd) {
		return;
	}

	// write the number of returned rows
	cachedestfd->write((char *)&totalrows,sizeof(long));

	// write the number of affected rows
	cachedestfd->write((char *)&affectedrows,sizeof(long));

	// write column descriptions to the cache file
	short	namelen;
	for (int i=0; i<colcount; i++) {
		namelen=strlen(columnarray[i]->name);
		cachedestfd->write((char *)&namelen,
					sizeof(short));
		cachedestfd->write(columnarray[i]->name,namelen);
		cachedestfd->write((char *)&columnarray[i]->type,
					sizeof(short));
		cachedestfd->write((char *)&columnarray[i]->length,
					sizeof(long));
	}

	// write the zero terminating the column descriptions
	short	zero=0;
	cachedestfd->write((char *)&zero,sizeof(short));
}

void	sqlrclient::cacheOutputBinds(int count) {

	if (resumed || !cachedestfd) {
		return;
	}

	// write the variable/value pairs to the cache file
	short	len;
	for (int i=0; i<count; i++) {
		len=strlen(outbindvars[i].variable);
		cachedestfd->write((char *)&len,sizeof(short));
		cachedestfd->write(outbindvars[i].variable,len);
		len=strlen(outbindvars[i].value.stringval);
		cachedestfd->write((char *)&len,sizeof(short));
		cachedestfd->write((char *)&outbindvars[i].value.stringval,len);
	}

	// write the -1 terminating the output binds
	short	minusone=-1;
	cachedestfd->write((char *)&minusone,sizeof(short));
}

int	sqlrclient::parseData(int rowtoget, int offset) {

	if (debug) {
		debugPrint("Parsing Data");
		debugPrint("\n");
	}

	// if we're already at the end of the result set, then just return
	if (endofresultset) {
		if (debug) {
			debugPrint("Already at the end of the result set");
			debugPrint("\n");
		}
		return 1;
	}

	// if we're reading a cache file, we may need to jump to a particular
	// place in the file
	jumpToOffset(offset);

	// if we're stepping through the result set, we can possibly 
	// skip a big chunk of it...
	skipRows(rowtoget);

	// tell the connection how many rows to send
	fetchRows();

	// useful variables
	long	length;
	char	*buffer=NULL;
	int	whichcol=0;

	// set firstrowindex to the index of the first row in the buffer
	firstrowindex=rowcount;

	// keep track of how large the buffer is
	int	rsbuffercount=0;

	// start at the first column
	currentcol=firstcol;

	// get rows
	for (;;) {
	
		// get the length of the field
		if (getLong(&length)!=sizeof(long)) {
			return -1;
		}

		// check for the end of the result set
		if (length==-1) {

			if (debug) {
				debugPrint("Got end of result set.");
				debugPrint("\n");
			}
			endofresultset=1;

			// if we were stepping through a cached result set
			// then we need to close the file
			clearCacheSource();
			break;
		}

		// if we're on the first column, start a new row,
		// reset the column pointer, and increment the
		// buffer counter and total row counter
		if (currentcol==firstcol) {

			if (!firstrow) {
				currentrow=new row();
				firstrow=currentrow;
			} else {
				currentrow->next=new row();
				currentrow=currentrow->next;
			}

			rsbuffercount++;
			rowcount++;
		}

		// if length==-3 then it's a long datatype... 
		// handle these a bit differently
		if (length==-3) {

			char	*oldbuffer;
			int	totallength=0;
			while (1) {

				// get the length of the chunk
				if (getLong(&length)!=sizeof(long)) {
					if (buffer) {
						delete buffer;
					}
					return -1;
				}

				// if length==-3 then we're done
				if (length==-3) {
					break;
				}

				buffer=new char[totallength+length+1];
				if (totallength) {
					strncpy(buffer,oldbuffer,totallength);
					delete oldbuffer;
					oldbuffer=buffer;
					buffer=buffer+totallength;
				} else {
					oldbuffer=buffer;
				}
				totallength=totallength+length;

				getString(buffer,length);
				buffer[length]=(char)NULL;
			}
			buffer=oldbuffer;
			length=totallength;

		} else if (length==NULL_VALUE_SIZE) {

			if (returnnulls) {
				buffer=NULL;
			} else {
				buffer=new char[1];
				buffer[0]=(char)NULL;
			}
			length=0;

		} else {

			// for non-long, non-NULL datatypes...
			// get the field into a buffer
			buffer=new char[length+1];
			getString(buffer,length);
			buffer[length]=(char)NULL;
		}

		// add the buffer to the current row
		currentrow->addField(buffer,length);
	
		if (debug) {
			if (buffer) {
				debugPrint("\"");
				debugPrint(buffer);
				debugPrint("\",");
			} else {
				debugPrint(buffer);
				debugPrint(",");
			}
		}

		// keep track of the longest field
		if (length>currentcol->longest) {
			currentcol->longest=length;
		}
		currentcol=currentcol->next;

		// move to the next column, handle end of row 
		if (!currentcol) {

			currentrow->noMoreFields();
			currentcol=firstcol;

			if (debug) {
				debugPrint("\n");
			}

			// check to see if we've gotten enough rows
			if (rsbuffersize && 
				rsbuffercount==rsbuffersize) {
				break;
			}
		}
	}

	// terminate the row list
	if (currentrow) {
		currentrow->next=NULL;
	}

	createFieldArray();

	// cache the rows
	cacheData();

	return 1;
}

void	sqlrclient::fetchRows() {

	if (debug) {
		debugPrint("Fetching ");
		debugPrint(rsbuffersize);
		debugPrint(" rows");
		debugPrint("\n");
	}

	// if we're reading from a cached result set, do nothing
	if (cachesourcefd>-1) {
		return;
	}

	// otherwise, send to the connection the number of rows to send back
	writeToServer((long)rsbuffersize);
}

void	sqlrclient::jumpToOffset(int offset) {

	if (offset && cachesourcefd>-1) {
		if (debug) {
			debugPrint("Jumping to offset: ");
			debugPrint((long)offset);
			debugPrint("\n");
		}
		lseek(cachesourcefd,offset,SEEK_SET);
	}
}

void	sqlrclient::skipRows(int rowtoget) {

	// calculate how many rows to skip
	long	skip=0;
	if (rsbuffersize) {
		skip=(long)((rowtoget-
				(rowtoget%rsbuffersize))-rowcount); 
		rowcount=rowcount+skip;
	}
	if (debug) {
		debugPrint("Skipping ");
		debugPrint((long)skip);
		debugPrint(" rows");
		debugPrint("\n");
	}

	// if we're reading from a cached result set we have to manually skip
	long	length;
	char	string;
	if (cachesourcefd>-1) {
		for (int i=0; i<skip*colcount; i++) {
			getLong(&length);
			// If length is -1 the we're at the end of the result
			// set.  We'll back up so that the -1 can be re-read by
			// parseData() which will abort accordingly.
			if (length==-1) {
				lseek(cachesourcefd,-1*sizeof(long),SEEK_CUR);
				return;
			}
			lseek(cachesourcefd,length,SEEK_CUR);
		}
		return;
	}

	// if we're reading from a connection, send the connection the 
	// number of rows to skip
	writeToServer((long)skip);
	return;
}

void	sqlrclient::cacheData() {

	if (!cachedestfd) {
		return;
	}

	// write the data to the cache file
	for (int i=0; i<rowcount-firstrowindex; i++) {
		for (int j=0; j<colcount; j++) {
			long	len;
			if (fieldarray[i][j]) {
				len=strlen(fieldarray[i][j]);
			} else {
				len=NULL_VALUE_SIZE;
			}
			cachedestfd->write((char *)&len,sizeof(long));
			
			if (len>0) {
				cachedestfd->write((char *)fieldarray[i][j],
						strlen(fieldarray[i][j]));
			}
		}
	}

	if (endofresultset) {
		finishCaching();
	}
}

void	sqlrclient::createColumnArrays() {

	if (debug) {
		debugPrint("Creating Column Arrays...");
		debugPrint("\n");
	}

	// build a 2d array of pointers to the column values
	// and a 2d array of pointers to the column names
	if (colcount) {
		columnarray=new column *[colcount+1];
		columnnamearray=new char *[colcount+1];
		currentcol=firstcol;
		for (int i=0; i<colcount; i++) {
			columnarray[i]=currentcol;
			columnnamearray[i]=currentcol->name;
			currentcol=currentcol->next;
		}
		columnarray[colcount]=NULL;
		columnnamearray[colcount]=NULL;
	}
}

void	sqlrclient::createFieldArray() {

	if (debug) {
		debugPrint("Creating Field Arrays...");
		debugPrint("\n");
	}

	// build a 3d array of pointers to the field values
	if (rowcount) {
		int	howmany=rowcount-firstrowindex;
		fieldarray=new char **[howmany+1];
		int i;
		for (i=0; i<howmany; i++) {
			fieldarray[i]=new char *[colcount+1];
			fieldarray[i][colcount]=NULL;
		}
		fieldarray[howmany]=NULL;
	
		currentrow=firstrow;
		field	*currentfield;
		for (i=0; i<howmany; i++) {
			currentfield=currentrow->first;
			for (int j=0; j<colcount; j++) {
				fieldarray[i][j]=currentfield->value;
				currentfield=currentfield->next;
			}
			currentrow=currentrow->next;
		}
	}
}

void	sqlrclient::handleError(char *err) {

	if (debug) {
		debugPrint("Handling Error");
		debugPrint("\n");
	}

	// If an error was passed in, use it. 
	// Otherwise, get the error from the server.
	if (err) {
		error=new char[strlen(err)+1];
		strcpy(error,err);
	} else {
		// get the length of the error string
		short	length;
		if (getShort(&length)!=sizeof(short)) {
			error=new char[55];
			strcpy(error,"There was an error, but the connection died trying to retrieve it.  Sorry.");
		} else {
			// get the error string
			error=new char[length+1];
			readFromServer(error,length);
			error[length]=(char)NULL;
		}
	}
	
	if (debug) {
		debugPrint(error);
		debugPrint("\n");
	}

	endofresultset=1;

	// handle cachefile
	if (resumed || !cachedestfd) {
		return;
	}

	// write the number of returned rows, affected rows 
	// and a zero to terminate the column descriptions
	int	longzero=0;
	short	shortzero=0;
	cachedestfd->write((char *)&longzero,sizeof(long));
	cachedestfd->write((char *)&longzero,sizeof(long));
	cachedestfd->write((char *)&shortzero,sizeof(short));
	finishCaching();
}

int	sqlrclient::firstRowIndex() {
	return firstrowindex;
}

int	sqlrclient::rowCount() {
	return rowcount;
}

int	sqlrclient::totalRows() {
	return totalrows;
}

int	sqlrclient::affectedRows() {
	return affectedrows;
}

int	sqlrclient::colCount() {
	return colcount;
}

char	*sqlrclient::errorMessage() {

	if (error) {
		return error;
	}
	return NULL;
}

void	sqlrclient::getNullsAsEmptyStrings() {
	returnnulls=0;
}

void	sqlrclient::getNullsAsNulls() {
	returnnulls=1;
}

char	*sqlrclient::getField(int row, int col) {

	if (rowcount && row>=0 && row>=firstrowindex && 
					col>=0 && col<colcount) {

		// in the event that we're stepping through the result set 
		// instead of buffering the entire thing, the requested row
		// may have to be fetched into the buffer...
		int	whichrow=fetchRowIntoBuffer(row);

		if (whichrow>-1) {
			return fieldarray[whichrow][col];
		}
	}
	return NULL;
}

char	*sqlrclient::getField(int row, char *col) {

	if (rowcount && row>=0 && row>=firstrowindex) {
		currentcol=firstcol;
		for (int i=0; i<colcount; i++) {
			if (!strcasecmp(currentcol->name,col)) {

				// in the event that we're stepping through the
				// result set instead of buffering the entire 
				// thing, the requested row may have to be 
				// fetched into the buffer...
				int	whichrow=fetchRowIntoBuffer(row);

				if (whichrow>-1) {
					return fieldarray[whichrow][i];
				}
				return NULL;
			}
			currentcol=currentcol->next;
		}
	}
	return NULL;
}

char	**sqlrclient::getRow(int row) {

	if (rowcount && row>=0 && row>=firstrowindex) {

		// in the event that we're stepping through the result set 
		// instead of buffering the entire thing, the requested row
		// may have to be fetched into the buffer...
		int	whichrow=fetchRowIntoBuffer(row);

		if (whichrow>-1) {
			return fieldarray[whichrow];
		}
	}
	return NULL;
}

int	sqlrclient::fetchRowIntoBuffer(int row) {

	// if we getting the entire result set at once, the the result set 
	// buffer index is the same as the requested row
	if (!rsbuffersize) {
		if (row<rowcount) {
			return row;
		}
		return -1;
	}

	// but, if we're not getting the entire result set at once
	// and if the requested row is not in the current range, 
	// fetch more data from the connection
	if (row>=(firstrowindex+rsbuffersize)) {
		if (connected || cachesourcefd>-1) {
			clearRowBuffers();
			parseData(row,0);
		} else {
			return -1;
		}
	}

	// return the buffer index corresponding to the requested row
	// or -1 if the requested row is past the end of the result set
	if (row<rowcount) {
		return row%rsbuffersize;
	}
	return -1;
}

char	**sqlrclient::getColumnNames() {

	return columnnamearray;
}

char	*sqlrclient::getColumnName(int col) {

	if (colcount && col>=0 && col<colcount) {
		return columnnamearray[col];
	}
	return NULL;
}

char	*sqlrclient::getColumnType(int col) {

	if (colcount && col>=0 && col<colcount) {
		return datatypestring[columnarray[col]->type];
	}
	return NULL;
}

int	sqlrclient::getColumnLength(int col) {

	if (colcount && col>=0 && col<colcount) {
		return columnarray[col]->length;
	}
	return 0;
}

char	*sqlrclient::getColumnType(char *col) {

	for (int i=0; i<colcount; i++) {
		if (!strcasecmp(columnarray[i]->name,col)) {
			return datatypestring[columnarray[i]->type];
		}
	}
	return NULL;
}

int	sqlrclient::getColumnLength(char *col) {

	for (int i=0; i<colcount; i++) {
		if (!strcasecmp(columnarray[i]->name,col)) {
			return columnarray[i]->length;
		}
	}
	return 0;
}

int	sqlrclient::getLongest(int col) {

	if (colcount && col>=0 && col<colcount) {
		return columnarray[col]->longest;
	} 
	return 0;
}

int	sqlrclient::getLongest(char *col) {

	for (int i=0; i<colcount; i++) {
		if (!strcasecmp(columnarray[i]->name,col)) {
			return columnarray[i]->longest;
		}
	}
	return 0;
}

void	sqlrclient::debugOn() {

	debug=1;
}

void	sqlrclient::debugOff() {

	debug=0;
}

int	sqlrclient::getDebug() {
	return debug;
}

void	sqlrclient::debugPrintFunction(int (*printfunction)(const char *,...)) {
	this->printfunction=printfunction;
}

void	sqlrclient::debugPrint(char *string) {
	if (printfunction) {
		(*printfunction)("%s",string);
	} else {
		cout << string;
	}
}

void	sqlrclient::debugPrint(long number) {
	if (printfunction) {
		(*printfunction)("%d",number);
	} else {
		cout << number;
	}
}

void	sqlrclient::debugPrint(double number) {
	if (printfunction) {
		(*printfunction)("%f",number);
	} else {
		cout << number;
	}
}

void	sqlrclient::copyReferences() {

	// set the flag
	copyrefs=1;

	// create a buffer to hold queries
	querybuffer=new char[MAXQUERYSIZE+1];
	queryptr=querybuffer;

	// make copies of some specific things
	char	*tempserver=new char[strlen(server)+1];
	char	*templistenerunixport=new char[strlen(listenerunixport)+1];
	char	*tempuser=new char[strlen(user)+1];
	char	*temppassword=new char[strlen(password)+1];

	strcpy(tempserver,server);
	strcpy(templistenerunixport,listenerunixport);
	strcpy(tempuser,user);
	strcpy(temppassword,password);

	server=tempserver;
	listenerunixport=templistenerunixport;
	user=tempuser;
	password=temppassword;

	// initialize the bind and substitution variables
	initVariables();
}
