#! /usr/bin/env python

# Copyright (c) 2000 Roman Milner
# See the file COPYING for more information

import socket
import string
import sys, os
import StringIO
import exceptions

class ListenerClient:

    def __init__(self, connection_parm, user, password):
        self.connection_parm = connection_parm
        self.user = user
        self.password = password
        self.box = connection_parm[0]
        self.orig_port = connection_parm[1]

    def connect(self):
        self.cur_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.cur_socket.connect(self.connection_parm[0], self.connection_parm[1])
        self.cur_socket.send('user:%spassword:%s' % (self.user, self.password))
        self.sock_file = self.cur_socket.makefile()
        self.port = self.__get_port(self.sock_file.read())
        self.cur_socket.close()
        return (self.port, self.box)
        
    def get_connection(self):
        return DataClient(self.box, self.orig_port, self)

    def __get_port(self, s):
        l = string.split(s, ':')
        return int(l[1][:-1])

class DataClient:

    str_type = type('')

    def __init__(self, box, orig_port, other):
        self.orig_port = orig_port
        self.other = other
        self.connect()
        
    def __del__(self):
        self.endSession()

    def connect(self, retry=1):
        port, box = self.other.connect()
        self.cur_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.cur_socket.connect(box, port)
            self.sock_file = self.cur_socket.makefile()
        except socket.error:
            if retry:
                # can't connect try 1 more time 
                self.connect(retry=0)
        self.__connected = 1
        
    def sendQuery(self, sql_string):
        try:
            self.cur_socket.send(sql_string)
        except socket.error:
            raise SQLRNotConnected
        self.__get_data()

    def rowCount(self):
        try:
            return len(self.__data)
        except AttributeError:
            return None

    def colCount(self):
        try:
            return len(self.__definitions)
        except AttributeError:
            return None

    def endSession(self):
        self.cur_socket.send('~')
        self.cur_socket.close()
        del(self.cur_socket)
        self.__connected = 0

    def getField(self, row, col):
        if type(col) == self.str_type:
            return self.__data[row][self.__col_by_name(col)]
        else:
            return self.__data[row][col]

    def errorMessage(self):
        return self.__cur_error

    def getRow(self, row_number):
        return self.__data[row_number]

    def getRowRange(self, beg, end):
        try:
            return self.__data[beg:end]
        except AttributeError:
            return None

    def getAll(self):
        return self.__data

    def getColumnName(self, col):
        return self.__definitions[col][0]

    def getColumnNames(self):
        rc = []
        for f in self.__definitions:
            rc.append(f[0])
        return rc

    def getColumnType(self, col):
        if type(col) == self.str_type:
            return self.__definitions[self.__col_by_name(col)][1]
        else:
            return self.__definitions[col][1]

    def getColumnLength(self, col):
        if type(col) == self.str_type:
            return self.__definitions[self.__col_by_name(col)][2]
        else:
            return self.__definitions[col][2]

    def __col_by_name(self, col_name):
        counter = 0
        for f in self.__definitions:
            if f[0] == col_name:
                return counter
            counter = counter + 1
        return None

    def __get_data(self):
        raw = self.__read_results()
        ds = self.__parse_results(raw)

    def __parse_results(self, raw):
        self.__data = []
        if raw[0] == 'E':
            self.__parse_error(raw)
        else:
            definition, data = string.split(raw, '\nD\n')
            self.__cur_error = None
            if len(definition) > 1:
                self.__parse_definition(definition)
                self.__data = []
                if len(data) > 1:
                    self.__parse_data(data)
            else:
                return

    def __parse_error(self, raw):
        self.__cur_error = raw

    def __parse_data(self, raw):
        raw = self.__quote_fix(raw)
        raw_list = string.split(raw, '"')
        del(raw)
        num_cols = self.colCount() + 2
        counter = 0
        self.__data = []
        cur_record = []
        cur_row = 0
        for l in raw_list:
            if counter <> 0 and counter < num_cols -1:
                cv = self.__unquote_fix(l)
                cv = self.__do_cast(cv, counter-1, cur_row)
                cur_record.append(cv)
            if counter >= num_cols -1:
                self.__data.append(cur_record)
                cur_record = []
                counter = 0
                cur_row = cur_row + 1
            counter = counter + 1

    def __do_cast(self, val, col, row):
        number_types = ['NUMBER', 'INT', 'SMALLINT']
        cur_type = self.getColumnType(col)
        f = val
        if val == '':
            return ''
        elif cur_type in number_types:
            if string.find(f, '.') <> -1:
                try:
                    return float(f)
                except ValueError:
                    return f
            else:
                try:
                    return int(f)
                except ValueError:
                    return f
        else:
            return f

    def __quote_fix(self, s):
        return string.replace(s, '\\"', "__\0qu||ote\0__")

    def __unquote_fix(self, s):
        return string.replace(s, "__\0qu||ote\0__", '"')
           
    def __parse_definition(self, raw):
        lines = StringIO.StringIO(raw).readlines()
        self.__definitions = []
        for l in lines[1:]:
            raw_fields = string.split(l, '"')
            self.__definitions.append((raw_fields[1],raw_fields[2],int(raw_fields[3])))

    def __read_results(self):
        rc = ''
        while 1:
	    try:
            	d = self.cur_socket.recv(1024)
	    except socket.error:
		return rc
            rc = rc + d
            if rc[-1:] == ";":
                break
        return rc


class Error(exceptions.StandardError):
    pass

class SQLRNotConnected(Error):
    pass
