# Copyright (c) 2000 Roman Milner
# See the file COPYING for more information

import CSQLRelay

class sqlrclient:

    """
    A wrapper for the sqlrelay client API.  Closely follows the C++ API.
    """

    str_type = type("")
    int_type = type(1)
    float_type = type(1.1)

    def __init__(self, connection_params, user, password, retrytime=1, tries=1):
        """ Opens a connections to the sqlrelay server and authenticates with
        user and password.  Failed connections are retried for tries times
        at retrytime interval.  connection_parames can either be a string for
        a local unix socket socket or a tuple of host and port number for tcp
        sockets."""
        if type(connection_params) == self.str_type:
            self.connection = CSQLRelay.connect("", 0, connection_params, user, password, retrytime, tries)
        else:
            self.connection = CSQLRelay.connect(connection_params[0], connection_params[1], "", user, password, retrytime, tries)
        self.__mod_ref = CSQLRelay
	
    def __del__(self):
        self.__mod_ref.endSession(self.connection)
        self.__mod_ref.free(self.connection)

    def setResultSetBufferSize(self, rows):
        """
        Sets the number of rows of the result set
        to buffer at a time.  0 (the default)
        means buffer the entire result set.
        """
        return CSQLRelay.setResultSetBufferSize(self.connection, rows)

    def getResultSetBufferSize(self):
        """
        Returns the number of result set rows that 
        will be buffered at a time or 0 for the
        entire result set.
        """
        return CSQLRelay.getResultSetBufferSize(self.connection)

    def ping(self):
        """
        Returns 1 if the database is up and 0
        if it's down.
        """
        return CSQLRelay.ping(self.connection)

    def identify(self):
        """
        Returns the type of database: 
          oracle7, oracle8, postgresql, mysql, etc.
        """
        return CSQLRelay.identify(self.connection)

    def cacheOn(self):
        """
        Sets query caching on.  Future queries
        will be cached to local files, the names of 
        which can be retrieved using 
        getCacheFileName() after each query is sent.
        
        A default time-to-live of 10 minutes is
        also set.
        """
        return CSQLRelay.cacheOn(self.connection)

    def setCacheTtl(self, ttl):
        """
        Sets the time-to-live for cached result
        sets. The sqlr-cachemanger will remove each 
        cached result set "ttl" seconds after it's 
        created.
        """
        return CSQLRelay.setCacheTtl(self.connection,ttl)

    def getCacheFileName(self):
        """
	Returns the name of the file containing the most
	recently cached result set.
        """
        return CSQLRelay.getCacheFileName(self.connection)

    def cacheOff(self):
        """
        Sets query caching off.
        """
        return CSQLRelay.cacheOff(self.connection)


    """
    If you don't need to use substitution or bind variables
    in your queries, use these two methods.
    """
    def sendQuery(self, query):
        """
        Send a SQL query to the server and
        gets a result set.
        """
        return CSQLRelay.sendQuery(self.connection, query)

    def sendFileQuery(self, path, file):
        """
        Send the SQL query in path/file to the server and
        gets a result set.
        """
        return CSQLRelay.sendFileQuery(self.connection, path, file)

    """
    If you need to use substitution or bind variables, in your
    queries use the following methods.  See the API documentation
    for more information about substitution and bind variables.
    """
    def prepareQuery(self, query):
        """
        Prepare to execute query.
        """
        return CSQLRelay.prepareQuery(self.connection, query)

    def prepareFileQuery(self, path, file):
        """
        Prepare to execute the contents of path/filename.
        """
        return CSQLRelay.prepareFileQuery(self.connection, path, file)

    def substitution(self, variable, value, precision=0, scale=0):
        """
        Define a substitution variable.
        """
        if value is None:
        	return CSQLRelay.subNone(self.connection, variable)
        elif type(value) == self.str_type:
        	return CSQLRelay.subString(self.connection, variable, value)
        elif type(value) == self.int_type:
        	return CSQLRelay.subLong(self.connection, variable, value)
        elif type(value) == self.float_type:
        	return CSQLRelay.subDouble(self.connection, variable, value,
							precision, scale)

    def inputBind(self, variable, value, precision=0, scale=0):
        """
        Define an input bind varaible.
        """
        if value is None:
        	return CSQLRelay.inputBindNone(self.connection, 
							variable)
        elif type(value) == self.str_type:
        	return CSQLRelay.inputBindString(self.connection, 
							variable, value)
        elif type(value) == self.int_type:
        	return CSQLRelay.inputBindLong(self.connection, 
							variable, value)
        elif type(value) == self.float_type:
        	return CSQLRelay.inputBindDouble(self.connection, 
							variable, value,
							precision, scale)

    def defineOutputBind(self, variable, length):
        """
        Define an output bind varaible.
        """
        return CSQLRelay.defineOutputBind(self.connection, variable, length)

    def substitutions(self, variables, values, precisions=None, scales=None):
        """
        Define substitution variables.
        """
	for i in range(len(variables)):
		if precisions and scales:
			substitution(variables[i],values[i],
					precisions[i],scales[i])
		else:
			substitution(variables[i],values[i])

    def inputBinds(self, variables, values, precisions=None, scales=None):
        """
        Define input bind variables.
        """
	for i in range(len(variables)):
		if precisions and scales:
			inputBind(variables[i],values[i],
					precisions[i],scales[i])
		else:
			inputBind(variables[i],values[i])

    def validateBinds(self):
        """
        If you are binding to any variables that 
        might not actually be in your query, call 
        this to ensure that the database won't try 
        to bind them unless they really are in the 
        query.
        """
        return CSQLRelay.validateBinds(self.connection)

    def executeQuery(self):
        """
        Execute the query that was previously
        prepared and bound.
        """
        return CSQLRelay.executeQuery(self.connection)

    def getOutputBind(self, variable):
        """
        Retrieve the value of an output bind variable.
        """
        return CSQLRelay.getOutputBind(self.connection, variable)

    def openCachedResultSet(self, filename):
        """
        Open a result set after a sendCachedQeury
        """
        return CSQLRelay.openCachedResultSet(self.connection, filename)

    def openCachedResultSetWithOffset(self, filename, offset, row):
        """
        Opens a cached result set and skips to
        "offset".  The "row" parameter tells the
        method which row that offset corresponds to.
        Returns 1 on success and 0 on failure.
        """
        return CSQLRelay.openCachedResultSetWithOffset(self.connection, filename, offset, row)

    def rowCount(self):
        """
        Returns the number of rows in the current result set.
        """
        return CSQLRelay.rowCount(self.connection)

    def colCount(self):
        """
        Returns the number of columns in the current result set.
        """
        return CSQLRelay.colCount(self.connection)

    def totalRows(self):
        """
        Returns the total number of rows that will 
        be returned in the result set.  Not all 
        databases support this call.  Don't use it 
        for applications which are designed to be 
        portable across databases.  -1 is returned
        by databases which don't support this option.
        """
        return CSQLRelay.totalRows(self.connection)

    def affectedRows(self):
        """
        Returns the number of rows that were 
        updated, inserted or deleted by the query.
        Not all databases support this call.  Don't 
        use it for applications which are designed 
        to be portable across databases.  -1 is 
        returned by databases which don't support 
        this option.
        """
        return CSQLRelay.affectedRows(self.connection)

    def firstRowIndex(self):
        """
        Returns the index of the first buffered row.
        This is useful when buffering only part of
        the result set at a time.
        """
        return CSQLRelay.firstRowIndex(self.connection)

    def errorMessage(self):
        """
        If the query failed, the error message will be returned
        from this method.  Otherwise, it returns None.
        """
        return CSQLRelay.errorMessage(self.connection)

    def endSession(self):
        """
        Ends the current session.
        """
        return CSQLRelay.endSession(self.connection)

    def suspendSession(self):
        """
        Disconnects this client from the current
        session but leaves the session open so 
        that another client can connect to it 
        using resumeSession().
        """
        return CSQLRelay.suspendSession(self.connection)

    def resumeSession(self, port, socket):
        """
        Resumes a session previously left open 
        using suspendSession().
        Returns 1 on success and 0 on failure.
        """
        return CSQLRelay.resumeSession(self.connection, port, socket)

    def resumeCachedSession(self, port, socket, cachefilename):
        """
        Resumes a session previously left open
        using suspendSession() and continues caching
        the result set to "cachefilename".
        Returns 1 on success and 0 on failure.
        """
        return CSQLRelay.resumeCachedSession(self.connection, port, socket, cachefilename)

    def getConnectionPort(self):
        """
        Returns the inet port that the client is 
        communicating over. This parameter may be 
        passed to another client for use in
        the resumeSession() method.
        """
        return CSQLRelay.getConnectionPort(self.connection)

    def getConnectionSocket(self):
        """
        Returns the unix socket that the client is 
        communicating over. This parameter may be 
        passed to another client for use in
        the resumeSession() method.
        """
        return CSQLRelay.getConnectionSocket(self.connection)

    def getCacheFileOffset(self):
        """
        Returns the offset in the currently open
        cache file.  This parameter may be used
        in the openCachedResultSet() method to
        jump to a particular offset in the file.
        """
        return CSQLRelay.getCacheFileOffset(self.connection)

    def getNullsAsEmptyStrings(self):
        """
        Tells the client to return NULL fields and output
        bind variables as empty strings.
        This is the default.
        """
	return CSQLRelay.getNullsAsEmptyStrings(self.connection)

    def getNullsAsNone(self):
        """
        Tells the client to return NULL fields and output
        bind variables as NULL's.
        """
	return CSQLRelay.getNullsAsNone(self.connection)

    def getField(self, row, col):
        """
        Returns the value of the specified row and
        column.  col may be a column name or number.
        """
        if type(col) == self.str_type:
            return CSQLRelay.getFieldByName(self.connection, row, col)
        else:
            return CSQLRelay.getFieldByIndex(self.connection, row, col)

    def getRow(self, row):
        """
        Returns a list of values in the given row.
        """
        return CSQLRelay.getRow(self.connection, row)

    def getRowRange(self, beg, end):
        """
        Returns a list of lists of the rows between beg and end.
        Note: this function has no equivalent in other SQL Relay API's.
        """
        max_rows = self.rowCount()
        if end >= max_rows:
            end = max_rows - 1
        return CSQLRelay.getRowRange(self.connection, beg, end)

    def getRowDictionary(self, row):
        """
        Returns the requested row as values in a dictionary
	with column names for keys.
        """
        return CSQLRelay.getRowDictionary(self.connection, row)

    def getColumnName(self, col):
        """
        Returns the name of column number col.
        """
        return CSQLRelay.getColumnName(self.connection, col)

    def getColumnNames(self):
        """
        Returns a list of column names in the current result set.
        """
        return CSQLRelay.getColumnNames(self.connection)

    def getColumnType(self, col):
        """
        Returns the type of the specified column.  col may
        be a name or number.
        """
        if type(col) == self.str_type:
            return CSQLRelay.getColumnTypeByName(self.connection, col)
        else:
            return CSQLRelay.getColumnTypeByIndex(self.connection, col)

    def getColumnLength(self, col):
        """
        Returns the length of the specified column.  col may
        be a name or number.
        """
        if type(col) == self.str_type:
            return CSQLRelay.getColumnLengthByName(self.connection, col)
        else:
            return CSQLRelay.getColumnLengthByIndex(self.connection, col)

    def getLongest(self, col):
        """
        Returns the length of the specified column.  col may
        be a name or number.
        """
        if type(col) == self.str_type:
            return CSQLRelay.getLongestByName(self.connection, col)
        else:
            return CSQLRelay.getLongestByIndex(self.connection, col)

    def debugOn(self):
        """
	Turn verbose debugging on.
        """
        return CSQLRelay.debugOn(self.connection);

    def debugOff(self):
        """
	Turn verbose debugging off.
        """
        return CSQLRelay.debugOff(self.connection);

    def getDebug(self):
        """
	Returns 1 if debugging is turned on and 0 if debugging is turned off.
        """
        return CSQLRelay.getDebug(self.connection);
