// Copyright (c) 1999-2000  David Muse
// See the file COPYING for more information

#include <mysqlconnection.h>

#include <datatypes.h>

#include <stdlib.h>
#include <string.h>

mysqlconnection::mysqlconnection() {
	connected=0;
	errmesg=NULL;
}

mysqlconnection::~mysqlconnection() {
	if (errmesg) {
		delete errmesg;
	}
}

int	mysqlconnection::getNumberOfConnectStringVars() {
	return NUM_CONNECT_STRING_VARS;
}

void	mysqlconnection::handleConnectString() {
	user=connectStringValue("user");
	password=connectStringValue("password");
	db=connectStringValue("db");
	host=connectStringValue("host");
	port=connectStringValue("port");
	socket=connectStringValue("socket");
}

int	mysqlconnection::logIn() {


	// handle host
	char	*hostval;
	if (host && host[0]) {
		hostval=host;
	} else {
		hostval="";
	}

	// handle port
	int	portval;
	if (port && port[0]) {
		portval=atoi(port);
	} else {
		portval=0;
	}

	// handle socket
	char	*socketval;
	if (socket && socket[0]) {
		socketval=socket;
	} else {
		socketval=NULL;
	}

	// handle db
	char	*dbval;
	if (db && db[0]) {
		dbval=db;
	} else {
		dbval="";
	}
	
	// initialize database connection structure
	if (!mysql_init(&mysql)) {
		return 0;
	}

	// log in
	if (!mysql_real_connect(&mysql,hostval,user,password,dbval,
					portval,socketval,0)) {
		logOut();
		return 0;
	} else {
		connected=1;
		return 1;
	}
}

void	mysqlconnection::logOut() {
	connected=0;
	mysql_close(&mysql);
}

int	mysqlconnection::executeQuery(char *query, long length) {

	// initialize counts
	ncols=0;
	nrows=0;

	// execute the query
	if (mysql_real_query(&mysql,query,length)) {
		return 0;
	}

	// store the result set
	if ((mysqlresult=mysql_store_result(&mysql))==(MYSQL_RES *)NULL) {

		// if there was an error then return failure, otherwise
		// the query must have been some DML or DDL
		char	*err=mysql_error(&mysql);
		if (err && err[0]) {
			return 0;
		} else {
			return 1;
		}
	}

	// get the column count
	ncols=mysql_num_fields(mysqlresult);

	// get the row count
	nrows=mysql_num_rows(mysqlresult);

	// get the row count
	affectedrows=mysql_affected_rows(&mysql);

	return 1;
}

char	*mysqlconnection::getErrorMessage(int *liveconnection) {

	// store the error message because mysql_ping will set it blank
	if (errmesg) {
		delete errmesg;
	}
	errmesg=new strstream();
	*errmesg << mysql_error(&mysql) << ends;

	// only return an error if the database is up
	if (connected && !mysql_ping(&mysql)) {
		*liveconnection=1;
		return errmesg->str();
	} else {
		*liveconnection=0;
		return "";
	}
}

void	mysqlconnection::returnHeader() {

	// send row counts
	sendRowCounts((long)nrows,(long)affectedrows);

	// for DML or DDL queries, return no column info
	if (!mysqlresult) {
		return;
	}

	// a useful variable
	int	type;

	// position ourselves at the first field
	mysql_field_seek(mysqlresult,0);

	// for each column...
	for (int i=0; i<ncols; i++) {

		// fetch the field
		mysqlfield=mysql_fetch_field(mysqlresult);

		// append column type to the header
		if (mysqlfield->type==FIELD_TYPE_STRING) {
			type=STRING_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_VAR_STRING) {
			type=VARSTRING_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_DECIMAL) {
			type=DECIMAL_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_TINY) {
			type=TINY_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_SHORT) {
			type=SHORT_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_LONG) {
			type=LONG_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_FLOAT) {
			type=FLOAT_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_DOUBLE) {
			type=DOUBLE_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_LONGLONG) {
			type=LONGLONG_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_INT24) {
			type=INT24_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_TIMESTAMP) {
			type=TIMESTAMP_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_DATE) {
			type=DATE_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_TIME) {
			type=TIME_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_DATETIME) {
			type=DATETIME_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_YEAR) {
			type=YEAR_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_NEWDATE) {
			type=NEWDATE_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_NULL) {
			type=NULL_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_ENUM) {
			type=ENUM_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_SET) {
			type=SET_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_TINY_BLOB) {
			type=TINY_BLOB_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_MEDIUM_BLOB) {
			type=MEDIUM_BLOB_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_LONG_BLOB) {
			type=LONG_BLOB_DATATYPE;
		} else if (mysqlfield->type==FIELD_TYPE_BLOB) {
			type=BLOB_DATATYPE;
		} else {
			type=UNKNOWN_DATATYPE;
		}

		// send column definition
		sendColumnDefinition(mysqlfield->name,
					strlen(mysqlfield->name),
					type,(int)mysqlfield->length);
	}
}

int	mysqlconnection::noRowsToReturn() {

	// for DML or DDL queries, return no data
	if (!mysqlresult) {
		return 1;
	}
	return 0;
}

int	mysqlconnection::fetchRow() {

	return ((mysqlrow=mysql_fetch_row(mysqlresult))!=NULL);
}

void	mysqlconnection::returnRow() {

	for (int col=0; col<ncols; col++) {

		if (mysqlrow[col]) {
			sendField(mysqlrow[col],strlen(mysqlrow[col]));
		} else {
			sendNullField();
		}
	}
}

void	mysqlconnection::cleanUpData() {
	if (mysqlresult!=(MYSQL_RES *)NULL) {
		mysql_free_result(mysqlresult);
	}
}

int	mysqlconnection::ping() {
	if (!mysql_ping(&mysql)) {
		return 1;
	}
	return 0;
}

char	*mysqlconnection::identify() {
	return "mysql";
}

int	mysqlconnection::isTransactional() {
	return 0;
}
