// Copyright (c) 1999-2000  David Muse
// See the file COPYING for more information

#include <sqliteconnection.h>

#include <datatypes.h>

#include <stdio.h>
#include <sys/ipc.h>
#include <sys/stat.h>

#ifdef __GNUC__
static	char	*sqliteconnection::query;
#endif

sqliteconnection::sqliteconnection() {
	errmesg=NULL;
	colnames=NULL;
	ncols=0;

	// create a semaphore set for thread synchronization
	int	vals[4]={0,0,0,0};
	sync=new semaphoreset(IPC_PRIVATE,4,vals,S_IRUSR|S_IWUSR);
}

sqliteconnection::~sqliteconnection() {
	if (errmesg) {
		delete errmesg;
	}
	delete sync;
}

int	sqliteconnection::getNumberOfConnectStringVars() {
	return NUM_CONNECT_STRING_VARS;
}

void	sqliteconnection::handleConnectString() {
	db=connectStringValue("db");
}

int	sqliteconnection::logIn() {

	if (errmesg) {
		delete errmesg;
		errmesg=NULL;
	}

	if (sqliteptr=sqlite_open(db,666,&errmesg)) {
		return 1;
	}
	return 0;
}

void	sqliteconnection::logOut() {
	sqlite_close(sqliteptr);
}

int	sqliteconnection::executeQuery(char *query, long length) {

	// re-init semaphores
	sync->setValue(0,0);
	sync->setValue(1,0);
	sync->setValue(2,0);
	sync->setValue(3,0);

	// Fork and detach a thread to do the actual query execution 
	// and callbacks.
	if (pthread_attr_init(&attr)) {
		return 0;
	}

	if (pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED)) {
		return 0;
	}

	this->query=query;
	if (pthread_create(&thread,&attr,
				(void * (*)(void *))&sqliteconnection::runQuery,
				(void *)this)) {
		return 0;
	}

	return 1;
}

void	sqliteconnection::runQuery(sqliteconnection *thisptr) {

	// set up some values
	thisptr->nomorerows=0;
	if (thisptr->errmesg) {
		delete thisptr->errmesg;
		thisptr->errmesg=NULL;
	}

	// execute the query here
	sqlite_exec(thisptr->sqliteptr,thisptr->query,
			(int(*)(void *,int,char **,char **))
				&sqliteconnection::handleRow,
			(void *)thisptr,&thisptr->errmesg);

	thisptr->nomorerows=1;

	// if no rows were sent, no header was sent either and the client
	// is waiting for the row counts, send them, 
	if (!thisptr->colnames) {
		thisptr->sendRowCounts((long)-1,(long)-1);
	}

	thisptr->sync->signal(1);
	thisptr->sync->signal(3);

	pthread_exit((void *)1);
}

char	*sqliteconnection::getErrorMessage(int *liveconnection) {
	*liveconnection=1;
	return errmesg;
}

int	sqliteconnection::handleRow(sqliteconnection *thisptr, 
					int argc, char **argv,
					char **columnnames) {

	if (!thisptr->colnames) {

		thisptr->sync->wait(0);

		// apparantly, we have to copy the column names array 
		// if we want to use it later
		thisptr->ncols=argc;
		thisptr->colnames=new char *[argc];
		for (int i=0; i<argc; i++) {
			thisptr->colnames[i]=new char[strlen(columnnames[i])+1];
			strcpy(thisptr->colnames[i],columnnames[i]);
		}

		thisptr->sync->signal(1);
	}

	thisptr->sync->wait(2);

	// send a row
	for (int col=0; col<argc; col++) {
		if (argv[col]) {
			thisptr->sendField(argv[col],strlen(argv[col]));
		} else {
			thisptr->sendNullField();
		}
	}

	thisptr->sync->signal(3);

	return 0;
}

void	sqliteconnection::returnHeader() {

	// signal the other thread to send a header, wait for it to finish
	if (!colnames) {
		sync->signal(0);
		sync->wait(1);
	}

	// actual and affected row counts are both unknown in sqlite
	sendRowCounts((long)-1,(long)-1);

	for (int i=0; i<ncols; i++) {

		// column type and size are unknown in sqlite
		sendColumnDefinition(colnames[i],
					strlen(colnames[i]),
					UNKNOWN_DATATYPE,0);
	}
}

int	sqliteconnection::noRowsToReturn() {
	return 0;
}

int	sqliteconnection::fetchRow() {

	// signal the other thread to fetch/send a row, wait for it to finish
	sync->signal(2);
	sync->wait(3);

	if (nomorerows) {
		return 0;
	}
	return 1;
}

void	sqliteconnection::returnRow() {
	// don't have to do anything here, the 
	// row gets sent when it gets fetched
}

void	sqliteconnection::cleanUpData() {
	if (colnames) {
		delete[] colnames;
		colnames=NULL;
		ncols=0;
	}
}

int	sqliteconnection::ping() {
	return 1;
}

char	*sqliteconnection::identify() {
	return "sqlite";
}

int	sqliteconnection::isTransactional() {
	return 0;
}
