// Copyright (c) 1999-2000  David Muse
// See the file COPYING for more information

#include <defaults.h>
#include <configfile.h>
#include <commandline.h>
#include <connection.h>

#include <iostream.h>
#include <iomanip.h>

#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <stdlib.h>
#include <time.h>
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif
#include <string.h>
#ifdef HAVE_STRINGS_H
	#include <strings.h>
#endif

#include <defines.h>


connection::connection() : daemon() {
	init=0;
	inbindcount=0;
	outbindcount=0;
	idmemory=NULL;
	sockseqname=NULL;
	unixsocket=NULL;
}

connection::~connection() {
	if (idmemory) {
		delete idmemory;
	}
	if (sockseqname) {
		delete sockseqname;
	}
	if (unixsocket) {
		delete unixsocket;
	}
}

void	connection::initConnection(int argc, char **argv) {

	init=1;

	// process the command line
	commandline	*cmdl=new commandline(argc,argv);

	// get the id from the command line
	char	*id=cmdl->value("-id");
	if (!id[0]) {
		id=DEFAULT_ID;
		cerr << "Warning: using default id." << endl;
	}

	// get the config file from the command line
	char	*config=cmdl->value("-config");
	if (!config[0]) {
		config=DEFAULT_CONFIG_FILE;
	}

	// get the connection id from the command line
	char	*connectionid=cmdl->value("-connectionid");

	// get the time to live from the command line
	ttl=atoi(cmdl->value("-ttl"));

	// parse the config file
	configfile		*cfgfile=NULL;
	connectstringnode	*connectstringlist=NULL;
	dynamicscaling=0;
	cfgfile=new configfile();
	if (cfgfile->parse(config,id)) {

		// run as user/group specified in the config file
		char	*runasuser=cfgfile->getRunAsUser();
		char	*runasgroup=cfgfile->getRunAsGroup();
		if (runasuser[0] && runasgroup[0]) {
			runAsUser(runasuser);
			runAsGroup(runasgroup);
		}

		connectstringlist=cfgfile->getConnectStrings();

		// commit or rollback at end of session?
		if (!strcmp(cfgfile->getEndOfSession(),"rollback")) {
			eoscommit=0;
		} else {
			eoscommit=1;
		}

		// how long before a suspended session times out
		sessiontimeout=cfgfile->getSessionTimeout();
		if (!sessiontimeout) {
			sessiontimeout=60;
		}

		// see if dynamic scaling will be used
		if (cfgfile->getMaxQueueLength>=0 &&
			cfgfile->getMaxConnections()>
				cfgfile->getConnections() &&
			cfgfile->getGrowBy()>0 && cfgfile->getTtl()>0) {
			dynamicscaling=1;
		}
	}

	// if a connectionid was supplied, find and use it's connect string
	connectstringnode	*currentnode;
	char	*connectstring=NULL;
	if (connectionid[0]) {
		currentnode=connectstringlist;
		while (currentnode) {
			if (!strcmp(connectionid,
					currentnode->getConnectionId())) {
				connectstring=currentnode->getString();
				break;
			}
			currentnode=currentnode->getNext();
		}
	}

	// if no connectstring was found or supplied, use the default
	if (!connectstring) {
		connectstring=DEFAULT_CONNECTSTRING;
		cerr << "Warning: using default connectstring." << endl;
	}

	// initialize the shared memory segment filename
	idfilename=new strstream();
	*idfilename << TMP_DIR << "/" << id << ends;
	int	idfd=open(idfilename->str(),O_CREAT|O_RDWR,S_IRUSR|S_IWUSR);
	close(idfd);

	// initialize shared memory segment for passing port
	idmemory=new sharedmemory(ftok(idfilename->str(),0),
				sizeof(unsigned int)+1024+
					(2*sizeof(unsigned short int)),
				S_IRUSR|S_IWUSR);

	// initialize the database up/down filename
	updown=new strstream();
	*updown << TMP_DIR << "/" << id << "-";
	*updown << connectionid << ends;

	// set up connect strings
	connectstringcount=getNumberOfConnectStringVars();
	connectstringvar=new strstream *[connectstringcount];
	connectstringval=new strstream *[connectstringcount];
	for (int i=0; i<connectstringcount; i++) {
		connectstringvar[i]=NULL;
		connectstringval[i]=NULL;
	}
	parseConnectString(connectstring);

	// set some useful variables
	handleConnectString();

	// initialize the announce semaphore
	int	vals[9]={1,1,0,0,1,1,0,0,0};
	semset=new semaphoreset(ftok(idfilename->str(),0),9,
				vals,S_IRUSR | S_IWUSR);

	// get the unix socket sequence file name
	sockseqname=new strstream();
	*sockseqname << TMP_DIR << "/sockseq" << ends;

	// clean up
	if (cfgfile) {
		delete cfgfile;
	}
	delete cmdl;

	// increment connection counter
	if (dynamicscaling) {
		incrementConnectionCount();
	}

	// detach from the controlling tty
	detach();
}

void	connection::closeConnection() {

	// decrement the connection counter
	if (dynamicscaling) {
		decrementConnectionCount();
	}

	// delete the shared memory segment name
	delete idfilename;

	// delete announce semaphore
	delete semset;

	// delete the updown file name
	delete updown;

	// clean up connect strings
	for (int i=0; i<connectstringcount; i++) {
		delete connectstringvar[i];
		delete connectstringval[i];
	}
	delete[] connectstringvar;
	delete[] connectstringval;

	// try to log out
	logOut();
}

void	connection::incrementConnectionCount() {

	// wait for access
	semset->wait(4);

	// increment the counter
	unsigned int	*connectioncount=(unsigned int *)idmemory->getPointer();
	(*connectioncount)++;

	// signal the scaler
	if (dynamicscaling) {
		semset->signal(8);
	}

	// signal that others may have access
	semset->signal(4);
}

void	connection::decrementConnectionCount() {

	// wait for access
	semset->wait(4);

	// decrement the counter
	unsigned int	*connectioncount=(unsigned int *)idmemory->getPointer();
	(*connectioncount)--;

	// signal that others may have access
	semset->signal(4);
}

void	connection::decrementSessionCount() {

	// wait for access
	semset->wait(5);

	// decrement the counter
	unsigned int	*sessioncount=
				(unsigned int *)(idmemory->getPointer()+
					sizeof(unsigned int));
	(*sessioncount)--;

	// signal that others may have access
	semset->signal(5);
}

void	connection::listen() {

	blockSignals();
	if (logIn()) {

		markDatabaseAvailable();
		for (;;) {

			waitForAvailableDatabase();

			openSockets();

			initSession();

			announceAvailability();

			while (suspendsession) {

				if (acceptConnection(accepttimeout,0)) {

					// have a session with the client
					clientSession();

				} else {

					// if the accept times out waiting for
					// someone to pick up the suspended
					// session, roll it back and kill it
					if (suspendsession) {
						if (isTransactional()) {
							rollback();
						}
						suspendsession=0;
					}
				}
			}

			closeServerSocket();

			if (dynamicscaling) {
				decrementSessionCount();
			}
		}
	} else {
		// display couldn't log in error
		int		placeholder;
		cerr << "Couldn't log into database, error was: " << endl;
		cerr << getErrorMessage(&placeholder) << endl;
	}
}

void	connection::blockSignals() {

	// the daemon class handles SIGTERM's and SIGINT's
	// and we need to handle SIGALRMS so dynamically spawned daemons
	// can shut down but we're going to block all other signals
	sigset_t	set;
	sigemptyset(&set);

	#ifdef HAVE_SIGHUP
		sigaddset(&set, SIGHUP);
	#endif
	#ifdef HAVE_SIGQUIT
		sigaddset(&set,SIGQUIT);
	#endif
	#ifdef HAVE_SIGILL
		sigaddset(&set,SIGILL);
	#endif
	#ifdef HAVE_SIGTRAP
		sigaddset(&set,SIGTRAP);
	#endif
	#ifdef HAVE_SIGABRT
		sigaddset(&set,SIGABRT);
	#endif
	#ifdef HAVE_SIGIOT
		sigaddset(&set,SIGIOT);
	#endif
	#ifdef HAVE_SIGBUS
		sigaddset(&set,SIGBUS);
	#endif
	#ifdef HAVE_SIGFPE
		sigaddset(&set,SIGFPE);
	#endif
	#ifdef HAVE_SIGUSR1
		sigaddset(&set,SIGUSR1);
	#endif
	#ifdef HAVE_SIGSEGV
		sigaddset(&set,SIGSEGV);
	#endif
	#ifdef HAVE_SIGUSR2
		sigaddset(&set,SIGUSR2);
	#endif
	#ifdef HAVE_SIGPIPE
		sigaddset(&set,SIGPIPE);
	#endif
	#ifdef HAVE_SIGSTKFLT
		sigaddset(&set,SIGSTKFLT);
	#endif
	#ifdef HAVE_SIGCHLD
		sigaddset(&set,SIGCHLD);
	#endif
	#ifdef HAVE_SIGCONT
		sigaddset(&set,SIGCONT);
	#endif
	#ifdef HAVE_SIGSTOP
		sigaddset(&set,SIGSTOP);
	#endif
	#ifdef HAVE_SIGTSTP
		sigaddset(&set,SIGTSTP);
	#endif
	#ifdef HAVE_SIGTTIN
		sigaddset(&set,SIGTTIN);
	#endif
	#ifdef HAVE_SIGTTOU
		sigaddset(&set,SIGTTOU);
	#endif
	#ifdef HAVE_SIGURG
		sigaddset(&set,SIGURG);
	#endif
	#ifdef HAVE_SIGXCPU
		sigaddset(&set,SIGXCPU);
	#endif
	#ifdef HAVE_SIGXFSZ
		sigaddset(&set,SIGXFSZ);
	#endif
	#ifdef HAVE_SIGVTALRM
		sigaddset(&set,SIGVTALRM);
	#endif
	#ifdef HAVE_SIGPROF
		sigaddset(&set,SIGPROF);
	#endif
	#ifdef HAVE_SIGWINCH
		sigaddset(&set,SIGWINCH);
	#endif
	#ifdef HAVE_SIGIO
		sigaddset(&set,SIGIO);
	#endif
	#ifdef HAVE_SIGPOLL
		sigaddset(&set,SIGPOLL);
	#endif
	#ifdef HAVE_SIGPWR
		sigaddset(&set,SIGPWR);
	#endif
	#ifdef HAVE_SIGUNUSED
		sigaddset(&set,SIGUNUSED);
	#endif
	#ifdef HAVE_SIGEMT
		sigaddset(&set,SIGEMT);
	#endif
	#ifdef HAVE_SIGSYS
		sigaddset(&set,SIGSYS);
	#endif
	#ifdef HAVE_SIGCLD
		sigaddset(&set,SIGCLD);
	#endif
	#ifdef HAVE_SIGWAITING
		sigaddset(&set,SIGWAITING);
	#endif
	#ifdef HAVE_SIGLWP
		sigaddset(&set,SIGLWP);
	#endif
	#ifdef HAVE_SIGFREEZE
		sigaddset(&set,SIGFREEZE);
	#endif
	#ifdef HAVE_SIGTHAW
		sigaddset(&set,SIGTHAW);
	#endif
	#ifdef HAVE_SIGCANCEL
		sigaddset(&set,SIGCANCEL);
	#endif
	#ifdef HAVE_SIGLOST
		sigaddset(&set,SIGLOST);
	#endif
	#ifdef HAVE__SIGRTMIN
		sigaddset(&set,_SIGRTMIN);
	#endif
	#ifdef HAVE__SIGRTMAX
		sigaddset(&set,_SIGRTMAX);
	#endif
	#ifdef HAVE_SIGRTMIN
		sigaddset(&set,SIGRTMIN);
	#endif
	#ifdef HAVE_SIGRTMAX
		sigaddset(&set,SIGRTMAX);
	#endif

	sigprocmask(SIG_SETMASK,&set,NULL);
}

void	connection::openSockets() {

	// get the next available unix socket and open it
	getUnixSocket();
	openUnixSocket(unixsocket->str(),
			S_IRUSR|S_IWUSR|S_IRGRP|
			S_IWGRP|S_IROTH|S_IWOTH,
			5);

	// open the next available inet socket
	openInetSocket(0,5);
}

void	connection::reLogIn() {

	// attempt to log in over and over, once every 5 seconds
	while (logIn()) {
		sleep(5);
	}
}

void	connection::announceAvailability() {

	// handle time-to-live
	alarm(ttl);

	// wait on the write mutex
	semset->wait(0);

	// cancel time-to-live alarm
	alarm(0);

	// write ports to the shared memory segment
	void	*idmemoryptr=idmemory->getPointer()+(2*sizeof(unsigned int));
	memcpy((void *)idmemoryptr,(void *)unixsocket->str(),
					unixsocket->pcount()-1);
	idmemoryptr=idmemoryptr+unixsocket->pcount()-1;
	*((char *)idmemoryptr)=':';
	idmemoryptr=idmemoryptr+1;
	*((unsigned short int *)idmemoryptr)=getInetServerPort();

	// signal that it's ok to read
	semset->signal(2);

	// wait for read to finish
	semset->wait(3);

	// signal on the write mutex
	semset->signal(0);
}

void	connection::clientSession() {

	if (suspendresultset) {
		// If the previous session suspended the result set, we send 
		// the client the index of the last row that was sent, the 
		// header, then the result set starting after the last row that
		// the previous session sent
		writeToClient(lastrow);
		returnResultSet();
	} else {
		// if the previous session did not suspend the result set,
		// just write a -1 out to tell the client that this is a new,
		// clean session
		writeToClient((long)-1);
	}

	// a session consists of getting a query and returning a result set
	// over and over
	suspendsession=0;
	while (1) {

		// get the query
		if (getQuery() && getInputBinds() && getOutputBinds()) {

			// handle ping or identify queries
			if (pingOrIdentify()) {
				continue;
			}

			// loop here to handle down databases
			while (1) {

				// process the query
				if (processQuery()) {

					// was the query a commit or rollback?
					commitOrRollback();

					// if the query processed ok then
					// return a result set
					returnResultSet();
					break;

				} else {

					// If the query didn't process ok,
					// handle the error.
					if (handleError()) {

						// if handleError returns a 0
						// then the error was a bad
						// query, otherwise it was
						// just a down database
						break;
					}
				}
			}

		} else {

			// if getting the query/bind variables failed,
			// close the connection
			closeClientSocket();
			if (!suspendsession) {
				endOfSession();
			}
			return;
		}
	}
}

int	connection::handleError() {

	// return the error unless the error was a dead connection, 
	// in which case, re-establish the connection
	if (!returnError()) {
		markDatabaseUnavailable();
		reLogIn();
		markDatabaseAvailable();
		return 0;
	}
	return 1;
}

int	connection::getQuery() {

	// get the length of the query
	if (readFromClient(&querylength)!=sizeof(long)) {
		return 0;
	}

	// bounds checking
	if (querylength>MAXQUERYSIZE) {
		return 0;
	}

	// read the query into the buffer
	if (readFromClient(querybuffer,querylength)!=querylength) {
		return 0;
	}
	querybuffer[querylength]=(char)NULL;

	// handle leave session open
	if (querybuffer[0]=='!') {
		suspendsession=1;
		accepttimeout=sessiontimeout;
		return 0;
	}

	// handle end of session
	if (querybuffer[0]=='~') {
		suspendsession=0;
		return 0;
	}

	return 1;
}


int	connection::getInputBinds() {

	// get the number of input bind variable/values
	if (readFromClient(&inbindcount)!=sizeof(short)) {
		return 0;
	}

	// bounds checking
	if (inbindcount>MAXVAR) {
		return 0;
	}
	
	short	inbindsize;
	short	type;
	if (inbindcount>0) {
		
		// fill the buffers
		for (int i=0; i<inbindcount && i<MAXVAR; i++) {

			// get the variable name size
			if (readFromClient(&inbindsize)!=sizeof(short)) {
				return 0;
			}

			// bounds checking
			if (inbindsize>BINDVARLENGTH) {
				return 0;
			}

			// get the variable name
			inbindvars[i].variablesize=inbindsize+1;
			inbindvars[i].variable[0]=':';
			if (readFromClient(inbindvars[i].variable+1,
						inbindsize)!=inbindsize) {
				return 0;
			}
			inbindvars[i].variable[inbindsize+1]=(char)NULL;

			// get the type
			if (readFromClient(&type)!=sizeof(short)) {
				return 0;
			}
			inbindvars[i].type=(bindtype)type;

			// get the value
			if (inbindvars[i].type==STRING) {

				if (readFromClient(&inbindsize)!=
							sizeof(short)) {
					return 0;
				}

				// bounds checking
				if (inbindsize>BINDVALUELENGTH) {
					return 0;
				}

				if (inbindsize==NULL_VALUE_SIZE) {
					inbindvars[i].value.
						stringval[0]=(char)NULL;
					inbindvars[i].valuesize=0;
					inbindvars[i].isnull=nullBindValue();
				} else {
					if (readFromClient(
						inbindvars[i].value.stringval,
						inbindsize)!=inbindsize) {
						return 0;
					}
					inbindvars[i].value.
						stringval[inbindsize]=
								(char)NULL;
					inbindvars[i].valuesize=inbindsize;
					inbindvars[i].isnull=nonNullBindValue();
				}
			} else if (inbindvars[i].type==LONG) {
				if (readFromClient(&inbindvars[i].value.
					longval)!=sizeof(long)) {
					return 0;
				}
			} else if (inbindvars[i].type==DOUBLE) {
				if (readFromClient(&inbindvars[i].value.
					doubleval.value)!=sizeof(double)) {
					return 0;
				}
				if (readFromClient(&inbindvars[i].value.
					doubleval.precision)!=sizeof(short)) {
					return 0;
				}
				if (readFromClient(&inbindvars[i].value.
					doubleval.scale)!=sizeof(short)) {
					return 0;
				}
			}
		}
	}

	return 1;
}


int	connection::getOutputBinds() {

	// get the number of output bind variable/values
	if (readFromClient(&outbindcount)!=sizeof(short)) {
		return 0;
	}

	// bounds checking
	if (outbindcount>MAXVAR) {
		return 0;
	}

	short	outbindsize;
	if (outbindcount>0) {
		
		// fill the buffers
		for (int i=0; i<outbindcount && i<MAXVAR; i++) {

			// get the variable name size
			if (readFromClient(&outbindsize)!=sizeof(short)) {
				return 0;
			}

			// bounds checking
			if (outbindsize>BINDVARLENGTH) {
				return 0;
			}

			// get the variable name
			outbindvars[i].variablesize=outbindsize+1;
			outbindvars[i].variable[0]=':';
			if (readFromClient(outbindvars[i].variable+1,
						outbindsize)!=outbindsize) {
				return 0;
			}
			outbindvars[i].variable[outbindsize+1]=(char)NULL;

			// get the size that the buffer to store the
			// result needs to be
			if (readFromClient(&outbindvars[i].valuesize)!=
							sizeof(short)) {
				return 0;
			}

			// stupid size checking
			if (outbindvars[i].valuesize>BINDVALUELENGTH) {
				return 0;
			}
		}
	}

	return 1;
}

int	connection::processQuery() {
	return (prepareQuery(querybuffer,querylength) && handleBinds() && 
			executeQuery(querybuffer,querylength));
}

int	connection::prepareQuery(char *query, long querylength) {
	// by default, do nothing...
	return 1;
}

int	connection::handleBinds() {
	
	// iterate through the arrays, binding values to variables
	for (int i=0; i<inbindcount; i++) {

		// bind the value to the variable
		if (inbindvars[i].type==STRING) {
			if (!inputBindValue(inbindvars[i].variable,
				inbindvars[i].variablesize,
				inbindvars[i].value.stringval,
				inbindvars[i].valuesize,
				&inbindvars[i].isnull)) {
				return 0;
			}
		} else if (inbindvars[i].type==LONG) {
			if (!inputBindValue(inbindvars[i].variable,
				inbindvars[i].variablesize,
				&inbindvars[i].value.longval)) {
				return 0;
			}
		} else if (inbindvars[i].type==DOUBLE) {
			if (!inputBindValue(inbindvars[i].variable,
				inbindvars[i].variablesize,
				&inbindvars[i].value.doubleval.value,
				inbindvars[i].value.doubleval.precision,
				inbindvars[i].value.doubleval.scale)) {
				return 0;
			}
		}
	}
	for (int i=0; i<outbindcount; i++) {

		// bind the value to the variable
		if (!outputBindValue(outbindvars[i].variable,
					outbindvars[i].variablesize,
					outbindvars[i].value.stringval,
					outbindvars[i].valuesize,
					&outbindvars[i].isnull)) {
			return 0;
		}
	}
	return 1;
}

int	connection::inputBindValue(char *variable, short variablesize,
					char *value, short valuesize,
					short *isnull) {
	// by default, do nothing...
	return 1;
}

int	connection::inputBindValue(char *variable, short variablesize,
					long *value) {
	// by default, do nothing...
	return 1;
}

int	connection::inputBindValue(char *variable, short variablesize,
					double *value, 
					short precision,
					short scale) {
	// by default, do nothing...
	return 1;
}

int	connection::outputBindValue(char *variable, short variablesize,
				char *value, short valuesize, short *isnull) {
	// by default, do nothing...
	return 1;
}

short	connection::nonNullBindValue() {
	return 0;
}

short	connection::nullBindValue() {
	return 1;
}

int	connection::bindValueIsNull(short isnull) {
	return 0;
}

int	connection::returnError() {

	// get the error message from the database
	// return value: 1 if database connection is still alive, 0 if not
	int		liveconnection;
	strstream	*error=new strstream();
	*error << getErrorMessage(&liveconnection);


	// only return an error message if the error wasn't a dead database
	if (liveconnection) {

		// send a -2 to indicate that there's an error
		writeToClient((long)-2);

		// send the error itself
		writeToClient((short)(error->rdbuf()->pcount()+
						strlen(querybuffer)+18));
		writeToClient(error->str(),error->rdbuf()->pcount());

		// send the attempted query back too
		writeToClient("\nAttempted Query:\n");
		writeToClient(querybuffer);
	}
	
	// clean up
	delete error;

	return liveconnection;
}

void	connection::returnResultSet() {

	// return the header
	returnHeader();

	// write a 0 to the client to terminate the header
	writeToClient((short)0);

	// return the output bind vars
	returnOutputBindValues();

	// write a -1 to the client to terminate the bind vars
	writeToClient((short)-1);

	// return the data
	if (returnData()) {
		// write a -1 to the client, signifying that the result set
		// has been sent
		writeToClient((long)-1);
	}
}

void	connection::returnOutputBindValues() {

	// run through the output bind values, sending them back
	for (int i=0; i<outbindcount; i++) {
		if (bindValueIsNull(outbindvars[i].isnull)) {
			writeToClient((short)NULL_VALUE_SIZE);
		} else {
			outbindvars[i].valuesize=
					strlen(outbindvars[i].value.stringval);
			writeToClient(outbindvars[i].valuesize);
			writeToClient(outbindvars[i].value.stringval,
						outbindvars[i].valuesize);
		}
	}
}

int	connection::returnData() {

	// get the number of rows to skip and fetch, client will always send
	// this, it has no idea what type of query it's sent
	long	skip;
	long	fetch;
	if ((skip=rowsToSkip())==-3) {
		return 0;
	}
	if ((fetch=rowsToFetch())==-3) {
		return 0;
	}

	// for some queries, there are no rows to return, 
	if (noRowsToReturn()) {
		cleanUpData();
		return 1;
	}

	// reinit suspendresultset and lastrow
	suspendresultset=0;
	lastrow=-1;

	// go fetch rows and columns
	for (;;) {

		// a -1 means retrieve no more rows
		if (skip==-1) {
			cleanUpData();
			return 1;
		}

		// a -2 means retrieve no more rows and suspend the result set
		if (skip==-2) {
			suspendresultset=1;
			return 1;
		}

		// skip the specified number of rows
		if (!skipRows(skip)) {
			return 1;
		}

		// send the specified number of rows back
		for (int i=0; (!fetch || i<fetch); i++) {

			if (!fetchRow()) {
				cleanUpData();
				return 1;
			}

			returnRow();

			lastrow++;
		}

		// get rows to skip and fetch for the next iteration
		if ((skip=rowsToSkip())==-3) {
			return 0;
		}
		if ((fetch=rowsToFetch())==-3) {
			return 0;
		}
	}
}

long	connection::rowsToFetch() {

	// get the number of rows to fetch 
	// 0 means fetch the rest, -1 means quit
	long	fetchcount;
	if (readFromClient(&fetchcount)!=sizeof(long)) {
		return -3;
	}
	return fetchcount;
}

long	connection::rowsToSkip() {

	// get the number of rows to skip
	long	skipcount;
	if (readFromClient(&skipcount)!=sizeof(long)) {
		return -3;
	}
	return skipcount;
}

int	connection::skipRows(int rows) {
	for (int i=0; i<rows; i++) {
		lastrow++;
		if (!fetchRow()) {
			return 0;
		}
	}
	return 1;
}

void	connection::parseConnectString(char *connectstring) {

	// handle connect string of form:
	//	 name1=value1;name2=value2;name3=value3;

	char	*ptr1=connectstring;
	char	*ptr2=connectstring;
	int	current=0;
	while (*ptr2) {
		while (*ptr2 && *ptr2!='=') {
			ptr2++;
		}
		if (!*ptr2) {
			break;
		}
		connectstringvar[current]=new strstream();
		while (ptr1!=ptr2) {
			*connectstringvar[current] << (char)*ptr1;
			ptr1++;
		}
		*connectstringvar[current] << ends;
		ptr1++;
		ptr2++;
		while (*ptr2 && *ptr2!=';') {
			ptr2++;
		}
		connectstringval[current]=new strstream();
		while (*ptr1 && ptr1!=ptr2) {
			*connectstringval[current] << (char)*ptr1;
			ptr1++;
		}
		*connectstringval[current] << ends;

		if (!*ptr2) {
			break;
		}

		current++;
		ptr1++;
		ptr2++;
	}
}

char	*connection::connectStringValue(char *variable) {

	// search for the given variable, return the given value,
	// search is case insensitive
	for (int i=0; i<connectstringcount; i++) {
		if (!connectstringvar[i]) {
			break;
		}
		if (!strcasecmp(variable,connectstringvar[i]->str())) {
			return connectstringval[i]->str();
		}
	}
	return NULL;
}

void	connection::sendRowCounts(long actual, long affected) {
	writeToClient(actual);
	writeToClient(affected);
}

void	connection::sendColumnDefinition(char *name, short namelen,
						short type, long size) {
	writeToClient(namelen);
	writeToClient(name,namelen);
	writeToClient(type);
	writeToClient(size);
}

void	connection::sendField(char *data, long size) {
	writeToClient(size);
	writeToClient(data,size);
}

void	connection::sendNullField() {
	writeToClient((long)NULL_VALUE_SIZE);
}

void	connection::startSendingLong() {
	writeToClient((long)-3);
}

void	connection::sendLongSegment(char *data, long size) {
	writeToClient(size);
	writeToClient(data,size);
}

void	connection::endSendingLong() {
	writeToClient((long)-3);
}

int	connection::setEnv(char *var, char *string) {

	#if defined(HAVE_PUTENV)
		strstream	pestr;
		pestr << var << "=" << string << ends;
		return putenv(pestr.str());
	#elif defined(HAVE_SETENV)
		return setenv(var,string,1);
	#else
		#error "Couldn't find a suitable replacement for putenv."
	#endif
}

char	*connection::getUnixSocket() {

	// open the sequence file and get the current port number
	fstream		sockseq(sockseqname->str(),
					ios::in|ios::out,
					S_IRUSR|S_IWUSR|S_IRGRP|
					S_IWGRP|S_IROTH|S_IWOTH);

	// prepare the buffer
	if (unixsocket) {
		delete unixsocket;
	}
	unixsocket=new strstream();

	// get the sequence number
	strstream	*tmp=new strstream();
	*tmp << sockseq.rdbuf() << ends;
	if (tmp->rdbuf()->pcount()>1) {
		*unixsocket << TMP_DIR << "/" << tmp->str() << ends;
	} else {
		*unixsocket << TMP_DIR << "/0" << ends;
	}

	// increment the sequence number
	int	unixsocketval=atoi(tmp->str())+1;
	if (unixsocketval>pow(2,32)) {
		unixsocketval=0;
	}
	sockseq.rdbuf()->seekpos(0);
	sockseq << unixsocketval;

	// clean up
	delete tmp;
	sockseq.close();
	return unixsocket->str();
}

void	connection::markDatabaseUnavailable() {

	// the database is down if the file isn't there
	unlink(updown->str());
}

void	connection::markDatabaseAvailable() {

	// the database is down if the file is there, 
	// opening and closing it will create it
	ofstream	updownfile(updown->str());
	updownfile.close();
}

void	connection::waitForAvailableDatabase() {

	// if the file is there, return, otherwise reconnect to the database
	ifstream	updownfile(updown->str());
	if (updownfile && updownfile.good()) {
		updownfile.close();
		return;
	}
	reLogIn();
	markDatabaseAvailable();
}

int	connection::queryIsNotSelect() {

	// scan the query, bypassing whitespace and comments.
	// look case insensitively for SELECT, return 0 if it's found,
	// 1 if it's not
	char	*ptr=querybuffer;
	while (*ptr && 
		(*ptr==' ' || *ptr=='\n' || *ptr=='	' || *ptr=='-')) {

		// skip any comments
		if (*ptr=='-') {
			while (*ptr && *ptr!='\n') {
				ptr++;
			}
		}
		ptr++;
	}

	// if the query is a select but not a select into then return false,
	// otherwise return true
	if (!strncasecmp(ptr,"select",6) && 
			strncasecmp(ptr,"select into ",12)) {
		return 0;
	}
	return 1;
}

int	connection::queryIsCommitOrRollback() {

	// scan the query, bypassing whitespace and comments.
	// look case insensitively for SELECT, return 0 if it's found,
	// 1 if it's not
	char	*ptr=querybuffer;
	while (*ptr && 
		(*ptr==' ' || *ptr=='\n' || *ptr=='	' || *ptr=='-')) {

		// skip any comments
		if (*ptr=='-') {
			while (*ptr && *ptr!='\n') {
				ptr++;
			}
		}
		ptr++;
	}

	// if the query is a commit or rollback, return true
	// otherwise return false
	if (!strncasecmp(ptr,"commit",6) || !strncasecmp(ptr,"rollback",8)) {
		return 1;
	}
	return 0;
}

int	connection::isTransactional() {
	return 1;
}

void	connection::commit() {
	executeQuery("commit",6);
}

void	connection::rollback() {
	executeQuery("rollback",8);
}

void	connection::initSession() {

	commitorrollback=0;
	suspendsession=1;
	suspendresultset=0;
	accepttimeout=5;
}

void	connection::commitOrRollback() {

	if (isTransactional()) {
		if (queryIsCommitOrRollback()) {
			commitorrollback=0;
		} else if (queryIsNotSelect()) {
			commitorrollback=1;
		}
	}
}

int	connection::pingOrIdentify() {

	// handle ping
	if (!strncmp(querybuffer,"ping",4)) {
		writeToClient((short)ping());
		return 1;
	}
		
	// handle identify
	if (!strncmp(querybuffer,"identify",8)) {
		char	*id=identify();
		short	idlen=(short)strlen(id);
		writeToClient(idlen);
		writeToClient(id,idlen);
		return 1;
	}

	return 0;
}

void	connection::endOfSession() {

	if (isTransactional()) {
		if (commitorrollback) {
			if (eoscommit) {
				commit();
			} else {
				rollback();
			}
		}
	}
}
