// Copyright (c) 1999-2000  David Muse
// See the file COPYING for more information

#include <defaults.h>
#include <configfile.h>
#include <commandline.h>
#include <listener.h>

#include <iostream.h>
#include <fstream.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif

listener::listener() : daemon(), server() {
	init=0;
	semset=NULL;
	idmemory=NULL;
	unixport=NULL;

	id=NULL;
	config=NULL;

	authcount=0;
	users=NULL;
	passwords=NULL;

	inetport=0;
}

listener::~listener() {
	if (init) {
		cleanUp();
	}
}

void	listener::initListener(int argc, char **argv) {

	init=1;

	// detach from the controlling tty
	detach();

	// read the commandline
	commandline	*cmdl=new commandline(argc,argv);

	// get the id
	char	*tmpid=cmdl->value("-id");
	if (!(tmpid && tmpid[0])) {
		tmpid=DEFAULT_ID;
		cerr << "Warning! using default id." << endl;
	}
	id=new char[strlen(tmpid)+1];
	strcpy(id,tmpid);

	// get the config file
	char	*tmpconfig=cmdl->value("-config");
	if (!(tmpconfig && tmpconfig[0])) {
		tmpconfig=DEFAULT_CONFIG_FILE;
	}
	config=new char[strlen(tmpconfig)+1];
	strcpy(config,tmpconfig);

	// parse the config file
	usernode		*usernodes=NULL;
	unsigned short int	port=0;
	dynamicscaling=0;
	configfile	*cfgfile=new configfile();
	if (cfgfile->parse(config,id)) {

		// run as user/group specified in the config file
		char	*runasuser=cfgfile->getRunAsUser();
		char	*runasgroup=cfgfile->getRunAsGroup();
		if (runasuser[0] && runasgroup[0]) {
			runAsUser(runasuser);
			runAsGroup(runasgroup);
		}

		// get the dynamic connection scaling parameters
		connections=cfgfile->getConnections();
		maxconnections=cfgfile->getMaxConnections();
		maxqueuelength=cfgfile->getMaxQueueLength();
		growby=cfgfile->getGrowBy();
		ttl=cfgfile->getTtl();

		// if dynamic scaling isn't going to be used, disable it
		if (maxqueuelength>=0 && maxconnections>connections &&
			growby>0 && ttl>0) {
			dynamicscaling=1;
		}
		

		// get users and port
		usernodes=cfgfile->getUsers();
		port=cfgfile->getPort();

		// get unixport
		char	*uport=cfgfile->getUnixPort();
		if (uport && uport[0]) {
			unixport=new char[strlen(uport)+1];
			strcpy(unixport,uport);
		}
	}


	// handle defaults...
	// if neither port nor unixport are specified, set up to 
	// listen on the default inet port
	if (!port && !unixport) {
		port=DEFAULT_PORT;
		cerr << "Warning! using default port." << endl;
	}

	// get the authentication strings
	usernode	*currentnode=usernodes;
	while (currentnode) {
		currentnode=currentnode->getNext();
		authcount++;
	}
	if (authcount==0) {
		users=new char *[2];
		passwords=new char *[2];

		users[0]=new char[strlen(DEFAULT_USER)+1];
		strcpy(users[0],DEFAULT_USER);

		passwords[0]=new char[strlen(DEFAULT_PASSWORD)+1];
		strcpy(passwords[0],DEFAULT_PASSWORD);

		authcount=1;
	} else {
		users=new char *[authcount];
		passwords=new char *[authcount];

		char	*user;
		char	*password;

		currentnode=usernodes;
		for (int i=0; i<authcount; i++) {

			user=currentnode->getUser();
			password=currentnode->getPassword();

			users[i]=new char[strlen(user)+1];
			strcpy(users[i],user);

			passwords[i]=new char[strlen(password)+1];
			strcpy(passwords[i],password);

			currentnode=currentnode->getNext();
		}
	}

	// init the unixports
	if (port) {
		openInetSocket(port,15);
	}
	if (unixport) {
		openUnixSocket(unixport,
				S_IRUSR|S_IWUSR|S_IRGRP|
				S_IWGRP|S_IROTH|S_IWOTH,
				15);
	}

	// initialize the signal handler variables
	deadchildhandler=NULL;

	// initialize the shared memory segment filename
	strstream	*idfilename=new strstream();
	*idfilename << TMP_DIR << "/" << id << ends;
	int	idfd=open(idfilename->str(),O_CREAT|O_RDWR,S_IRUSR|S_IWUSR);
	close(idfd);

	// connect to the shared memory segment
	idmemory=new sharedmemory(ftok(idfilename->str(),0),
				sizeof(unsigned int)+1024+
					(2*sizeof(unsigned short int)),
				S_IRUSR|S_IWUSR);

	// create (or connect) to the semaphore set
	int	vals[9]={1,1,0,0,1,1,0,0,0};
	semset=new semaphoreset(ftok(idfilename->str(),0),9,
					vals,S_IRUSR | S_IWUSR);

	// clean up
	delete idfilename;
	delete cfgfile;
	delete cmdl;
}

void	listener::cleanUp() {

	if (deadchildhandler) {
		delete deadchildhandler;
	}

	for (int i=0; i<authcount; i++) {
		delete[] users[i];
		delete[] passwords[i];
	}
	if (users) {
		delete[] users;
	}
	if (passwords) {
		delete[] passwords;
	}

	if (semset) {
		delete semset;
	}

	if (idmemory) {
		delete idmemory;
	}

	if (unixport) {
		delete[] unixport;
	}

	if (id) {
		delete[] id;
	}

	if (config) {
		delete[] config;
	}
}

void	listener::listen() {

	blockSignals();

	// handle signals
	handleDeadChild();

	for(;;) {

		// accept a tcp connection
		acceptConnection(-1,-1);

		// fork a child
		if (!fork()) {

			// close the duplicated server unixport connection
			closeServerSocket();

			// handle authentication
			int	authstatus=getAuth();

			// 3 possible outcomes: 1=pass 0=fail -1=bad data
			if (authstatus==1) {

				if (dynamicscaling) {
					incrementSessionCount();
				}
				getAConnection();
				writeToClient(unixportstrlen);
				writeToClient(unixportstr);
				writeToClient((short)inetport);

			} else if (authstatus==0) {
				returnAuthError();
			}

			// close the duplicated connection back to the client
			closeClientSocket();

			// clean up and exit
			cleanUp();
			_exit(0);
		}

		// close the connection back to the client
		closeClientSocket();
	}
}

void	listener::blockSignals() {

	// the daemon class handles SIGTERM's and SIGINT's and 
	// we're handling SIGCHLDS
	// so we're going to block all other signals 
	sigset_t	set;
	sigemptyset(&set);

	#ifdef HAVE_SIGHUP
		sigaddset(&set, SIGHUP);
	#endif
	#ifdef HAVE_SIGQUIT
		sigaddset(&set,SIGQUIT);
	#endif
	#ifdef HAVE_SIGILL
		sigaddset(&set,SIGILL);
	#endif
	#ifdef HAVE_SIGTRAP
		sigaddset(&set,SIGTRAP);
	#endif
	#ifdef HAVE_SIGABRT
		sigaddset(&set,SIGABRT);
	#endif
	#ifdef HAVE_SIGIOT
		sigaddset(&set,SIGIOT);
	#endif
	#ifdef HAVE_SIGBUS
		sigaddset(&set,SIGBUS);
	#endif
	#ifdef HAVE_SIGFPE
		sigaddset(&set,SIGFPE);
	#endif
	#ifdef HAVE_SIGUSR1
		sigaddset(&set,SIGUSR1);
	#endif
	#ifdef HAVE_SIGSEGV
		sigaddset(&set,SIGSEGV);
	#endif
	#ifdef HAVE_SIGUSR2
		sigaddset(&set,SIGUSR2);
	#endif
	#ifdef HAVE_SIGPIPE
		sigaddset(&set,SIGPIPE);
	#endif
	#ifdef HAVE_SIGALRM
		sigaddset(&set,SIGALRM);
	#endif
	#ifdef HAVE_SIGSTKFLT
		sigaddset(&set,SIGSTKFLT);
	#endif
	#ifdef HAVE_SIGCONT
		sigaddset(&set,SIGCONT);
	#endif
	#ifdef HAVE_SIGSTOP
		sigaddset(&set,SIGSTOP);
	#endif
	#ifdef HAVE_SIGTSTP
		sigaddset(&set,SIGTSTP);
	#endif
	#ifdef HAVE_SIGTTIN
		sigaddset(&set,SIGTTIN);
	#endif
	#ifdef HAVE_SIGTTOU
		sigaddset(&set,SIGTTOU);
	#endif
	#ifdef HAVE_SIGURG
		sigaddset(&set,SIGURG);
	#endif
	#ifdef HAVE_SIGXCPU
		sigaddset(&set,SIGXCPU);
	#endif
	#ifdef HAVE_SIGXFSZ
		sigaddset(&set,SIGXFSZ);
	#endif
	#ifdef HAVE_SIGVTALRM
		sigaddset(&set,SIGVTALRM);
	#endif
	#ifdef HAVE_SIGPROF
		sigaddset(&set,SIGPROF);
	#endif
	#ifdef HAVE_SIGWINCH
		sigaddset(&set,SIGWINCH);
	#endif
	#ifdef HAVE_SIGIO
		sigaddset(&set,SIGIO);
	#endif
	#ifdef HAVE_SIGPOLL
		sigaddset(&set,SIGPOLL);
	#endif
	#ifdef HAVE_SIGPWR
		sigaddset(&set,SIGPWR);
	#endif
	#ifdef HAVE_SIGUNUSED
		sigaddset(&set,SIGUNUSED);
	#endif
	#ifdef HAVE_SIGEMT
		sigaddset(&set,SIGEMT);
	#endif
	#ifdef HAVE_SIGSYS
		sigaddset(&set,SIGSYS);
	#endif
	#ifdef HAVE_SIGWAITING
		sigaddset(&set,SIGWAITING);
	#endif
	#ifdef HAVE_SIGLWP
		sigaddset(&set,SIGLWP);
	#endif
	#ifdef HAVE_SIGFREEZE
		sigaddset(&set,SIGFREEZE);
	#endif
	#ifdef HAVE_SIGTHAW
		sigaddset(&set,SIGTHAW);
	#endif
	#ifdef HAVE_SIGCANCEL
		sigaddset(&set,SIGCANCEL);
	#endif
	#ifdef HAVE_SIGLOST
		sigaddset(&set,SIGLOST);
	#endif
	#ifdef HAVE__SIGRTMIN
		sigaddset(&set,_SIGRTMIN);
	#endif
	#ifdef HAVE__SIGRTMAX
		sigaddset(&set,_SIGRTMAX);
	#endif
	#ifdef HAVE_SIGRTMIN
		sigaddset(&set,SIGRTMIN);
	#endif
	#ifdef HAVE_SIGRTMAX
		sigaddset(&set,SIGRTMAX);
	#endif

	sigprocmask(SIG_SETMASK,&set,NULL);
}

void	listener::handleDeadChild() {
	deadchildhandler=new signalhandler();
	deadchildhandler->setHandler((void *)waitOnChildren);
	deadchildhandler->addFlag(SA_NOCLDSTOP);
	deadchildhandler->setSignal(SIGCHLD);
}

int	listener::getAuth() {

	// Get the user/password. For either one, if they are too big or
	// if there's a read error, just exit with an error code
	long	size;
	readFromClient(&size);
	if (size>USERSIZE || readFromClient(userbuffer,size)<size) {
		return -1;
	}
	readFromClient(&size);
	if (size>USERSIZE || readFromClient(passwordbuffer,size)<size) {
		return -1;
	}

	// return 1 if what the client sent matches one of the user/password
	// sets and 0 if no match is found
	for (int i=0; i<authcount; i++) {
		if (!strcmp(userbuffer,users[i]) &&
			!strcmp(passwordbuffer,passwords[i])) {
			return 1;
		}
	}
	return 0;
}

void	listener::getAConnection() {

	// wait on the read mutex
	semset->wait(1);

	// buffer to hold the port
	unsigned short int	connectionport;

	// wait on a writer
	semset->wait(2);

	// get the unix port
	unixportstrlen=0;
	char	*ptr=(char *)(idmemory->getPointer()+(2*sizeof(unsigned int)));
	while (*ptr!=':' && unixportstrlen<MAXPATHLEN) {
		unixportstr[unixportstrlen]=*ptr;
		unixportstrlen++;
		ptr++;
	}
	unixportstr[unixportstrlen]=(char)NULL;
	ptr++;

	// get the inet port
	inetport=*((unsigned short int *)ptr);

	// done reading
	semset->signal(3);

	// signal on the read mutex
	semset->signal(1);
	
}

void	listener::incrementSessionCount() {

	// wait for access
	semset->wait(5);

	// increment the counter
	unsigned int	*sessioncount=
				(unsigned int *)(idmemory->getPointer()+
					sizeof(unsigned int));
	(*sessioncount)++;

	if (dynamicscaling) {
		// signal the scaler
		semset->signal(6);

		// wait for the scaler
		semset->wait(7);
	}

	// signal that others may have access
	semset->signal(5);
}

void	listener::returnAuthError() {

	// sleep before and after returning an
	// authentication error to discourage
	// brute-force password attacks
	sleep(2);
	writeToClient((short)-2);
	writeToClient((short)23);
	writeToClient("\nAuthentication Error.\n");
	sleep(2);
}
