// Copyright (c) 1999-2000  David Muse
// See the file COPYING for more information

#include <defaults.h>
#include <commandline.h>
#include <scaler.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ipc.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#ifdef HAVE_UNISTD_H
	#include <unistd.h>
#endif

scaler::scaler() : daemon() {
	init=0;
	semset=NULL;
	idmemory=NULL;
	cfgfile=NULL;

	id=NULL;
	config=NULL;
	dbase=NULL;

	randnum=NULL;
}

scaler::~scaler() {
	if (init) {
		cleanUp();
	}
}

void	scaler::initScaler(int argc, char **argv) {

	init=1;

	// detach from the controlling tty
	detach();

	// read the commandline
	commandline	*cmdl=new commandline(argc,argv);

	// get the id
	char	*tmpid=cmdl->value("-id");
	if (!(tmpid && tmpid[0])) {
		tmpid=DEFAULT_ID;
		cerr << "Warning! using default id." << endl;
	}
	id=new char[strlen(tmpid)+1];
	strcpy(id,tmpid);

	// get the config file
	char	*tmpconfig=cmdl->value("-config");
	if (!(tmpconfig && tmpconfig[0])) {
		tmpconfig=DEFAULT_CONFIG_FILE;
	}
	config=new char[strlen(tmpconfig)+1];
	strcpy(config,tmpconfig);

	// parse the config file
	cfgfile=new configfile();
	if (cfgfile->parse(config,id)) {

		// run as user/group specified in the config file
		char	*runasuser=cfgfile->getRunAsUser();
		char	*runasgroup=cfgfile->getRunAsGroup();
		if (runasuser[0] && runasgroup[0]) {
			runAsUser(runasuser);
			runAsGroup(runasgroup);
		}

		// get the dynamic connection scaling parameters
		connections=cfgfile->getConnections();
		maxconnections=cfgfile->getMaxConnections();
		maxqueuelength=cfgfile->getMaxQueueLength();
		growby=cfgfile->getGrowBy();
		ttl=cfgfile->getTtl();

		// get the database type
		dbase=new char[strlen(cfgfile->getDbase())+1];
		strcpy(dbase,cfgfile->getDbase());

		// get the list of connect strings
		connectstringlist=cfgfile->getConnectStrings();

		// add up the connection metrics
		connectstringnode	*currentnode=connectstringlist;
		metrictotal=0;
		while (currentnode) {
			metrictotal=metrictotal+currentnode->getMetric();
			currentnode=currentnode->getNext();
		}
	}

	// initialize the shared memory segment filename
	idfilename=new strstream();
	*idfilename << TMP_DIR << "/" << id << ends;
	int	idfd=open(idfilename->str(),O_CREAT|O_RDWR,S_IRUSR|S_IWUSR);
	close(idfd);

	// create (or connect) to the semaphore set
	int	vals[9]={1,1,0,0,1,1,0,0,0};
	semset=new semaphoreset(ftok(idfilename->str(),0),9,
					vals,S_IRUSR | S_IWUSR);

	// connect to the shared memory segment
	idmemory=new sharedmemory(ftok(idfilename->str(),0),
				sizeof(unsigned int)+1024+
					(2*sizeof(unsigned short int)),
				S_IRUSR|S_IWUSR);

	// set up random number generator
	randnum=new randomnumber();
	currentseed=time(NULL);

	// clean up
	delete cmdl;
}

void	scaler::cleanUp() {

	delete idfilename;

	if (semset) {
		delete semset;
	}

	if (idmemory) {
		delete idmemory;
	}

	if (id) {
		delete id;
	}

	if (config) {
		delete config;
	}

	if (dbase) {
		delete dbase;
	}

	if (cfgfile) {
		delete cfgfile;
	}

	if (randnum) {
		delete randnum;
	}
}


void	scaler::openMoreConnections() {

	// wait for a new listener to fire up and increment the listener count
	semset->wait(6);

	// maxqueuelength<0, maxconnections<=connections, growby<=0 or ttl<=0
	// would disable this whole feature (or cause problems) so
	// catch them up front
	if (maxqueuelength<0 || maxconnections<=connections || 
		growby<=0 || ttl<=0) {
		semset->signal(7);
		return;
	}

	int	sessions=countSessions();
	int	connections=countConnections();

	// signal listener to keep going
	semset->signal(7);

	// do we need to open more connections?
	if ((sessions-connections)<=maxqueuelength) {
		return;
	}

	// can more be opened, or will we exceed the max?
	if ((connections+growby)<=maxconnections) {

		// open "growby" connections
		for (int i=0; i<growby; i++) {

			getRandomConnectionId();

			// build the command to start a connection
			strstream	*command=new strstream();
			*command << "sqlr-connection-" << dbase;
			*command << " -ttl " << ttl;
			*command << " -id " << id;
			*command << " -connectionid " << connectionid;
			*command << " -config " << config;
			*command << ends;

			// start another connection
			system(command->str());

			// clean up
			delete command;

			// wait for the connection count to increase
			semset->wait(8);
		}
	}
}

void	scaler::getRandomConnectionId() {

	// get a scaled random number
	randnum->setSeed(currentseed);
	randnum->generateNumber();
	currentseed=randnum->getNumber();
	int	scalednum=randnum->getScaledNumber(0,metrictotal);

	// run through list, decrementing scalednum by the metric
	// for each, when scalednum is 0, pick that connection id
	connectstringnode	*currentnode=connectstringlist;
	while (currentnode) {
		scalednum=scalednum-currentnode->getMetric();
		if (scalednum<=0) {
			connectionid=currentnode->getConnectionId();
			break;
		}
		currentnode=currentnode->getNext();
	}
}

int	scaler::countSessions() {

	// get the number of open connections
	unsigned int	*sessions=
				(unsigned int *)(idmemory->getPointer()+
					sizeof(unsigned int));

	return (int)(*sessions);
}

int	scaler::countConnections() {

	// wait for access to the connection counter
	semset->wait(4);

	// get the number of connections
	unsigned int	*connections=
				(unsigned int *)idmemory->getPointer();

	// signal that the connection counter may be accessed by someone else
	semset->signal(4);

	return (int)(*connections);
}

void	scaler::loop() {

	for (;;) {
		// open more connections if necessary
		openMoreConnections();
	}
}
